package src.BAALL

import java.text.SimpleDateFormat
import java.util
import scala.collection.parallel.mutable
import src.ship._

/**
 * Created with IntelliJ IDEA.
 * User: autexier
 * Date: 11.10.2013
 * Time: 12:37
 */
case class BAALLButtonEvent(id: String, target: String, value: String, time: Long) extends JavaLogging {
  def toOntology(ontology:Ontology,previous: Option[BAALLButtonEvent]): List[ABoxFormula] = {
    val dateFormat = new SimpleDateFormat("yyyy-MM-dd'T'HH:mm:ssZ") //1956-06-25T04:00:00-0500
    val cal = util.Calendar.getInstance()

    var datetime: String = dateFormat.format(time) // 1956-06-25T04:00:00-0500
    val (head, tail) = datetime.splitAt(datetime.length - 2) // 1956-06-25T04:00:00-05:00
    datetime = head + ":" + tail
    //println(datetime)
    val islight = ontology.query(":"+target+":Door").isEmpty
    var proper: Boolean = true
    val button_ontology_ind = Const("button_" + id)
    val buttonevent_ontology_ind = Const("buttonevent_" + time)
    val button_value_ind = value match {
      case "on" => if (islight) Const("switchon") else Const("doopen")
      case "off" => if (islight) Const("switchoff") else Const("doclose")
      case _ =>
        log.warning("unexpected button value: " + value)
        proper = false
        Const(value)
    }
    (previous match {
      case Some(b) => List(NamedRole("previous_buttonevent").roleAssert(buttonevent_ontology_ind, Const("buttonevent_" + b.time)))
      case _ => Nil
    }) ++ (if (proper)
      List(
        ConceptConstant("ButtonEvent").mkAssertion(buttonevent_ontology_ind),
        NamedRole("from_button").roleAssert(buttonevent_ontology_ind,button_ontology_ind),
        NamedRole("button_target").roleAssert(buttonevent_ontology_ind, Const(target)),
        NamedRole("button_value").roleAssert(buttonevent_ontology_ind, button_value_ind),
        DataRole("button_time").roleAssert(buttonevent_ontology_ind, Const(datetime))
      )
    else Nil)
  }
}

class BAALLButtons {
  val mapping = new mutable.ParHashMap[String, BAALLButtonEvent]()

  def insertButton(button: BAALLButtonEvent) {
    println("Inserting button "+button)
    mapping update(button.target, button)
  }

  def getButtons: List[BAALLButtonEvent] = {
    mapping.values.toList.sortWith((b1, b2) => b1.time > b2.time)
  }

  def getPreviousButton(buttonname: String): Option[BAALLButtonEvent] = {
    val buttonsit = getButtons.iterator
    var result: Option[BAALLButtonEvent] = None
    while (buttonsit.hasNext && buttonsit.next().id != buttonname) {}
    if (buttonsit.hasNext) result = Some(buttonsit.next())
    result
  }
}




