package src.BAALL

import akka.actor.ActorDSL._
import akka.actor.{ActorLogging, ReceiveTimeout}
import java.io.FileNotFoundException
import org.xml.sax.SAXParseException
import scala.concurrent.duration._
import scala.xml.Node
import src.ship.JavaLogging

/**
 * Created with IntelliJ IDEA.
 * User: dodi
 * Date: 23.01.13
 * Time: 15:29
 */
class BaalLightControl {
}

case class RGBVal(r: Int = 0, g: Int = 0, b: Int = 0,id:Option[Int]) extends JavaLogging {
  override def toString: String = {
    "{ \"r\": " + r + ", \"g\": " + g + ", \"b\": " + b +
      (if (!id.isEmpty) ", id:" + id.get else "") + " }"
  }

  def setLight(url: String, argmap: Map[String, String]): Boolean = {
    var first = true
    try {
      id match {
        case Some(i) =>
          Http(url)
            .param("id",i.toString)
            .param(argmap.getOrElse("r", argmap.getOrElse("r", "1")), r.toString)
            .param(argmap.getOrElse("g", argmap.getOrElse("g", "2")), g.toString)
            .param(argmap.getOrElse("b", argmap.getOrElse("b", "3")), b.toString).asString
        case None =>
      Http(url)
        .param(argmap.getOrElse("r", argmap.getOrElse("r", "1")), r.toString)
        .param(argmap.getOrElse("g", argmap.getOrElse("g", "2")), g.toString)
        .param(argmap.getOrElse("b", argmap.getOrElse("b", "3")), b.toString).asString
      }
      first = false
    }
    catch {
      case e: Throwable => log.info("setLight " + url + " " + e)
    }
    first
  }
}

class LightMode(vals: Seq[RGBVal]) {
  var pos = 0
  var first = true
  val len = vals.length

  def step(url: String, argmap: Map[String, String]) {
    if (pos < len)
      if (first || len > 1) first = vals.apply(pos).setLight(url, argmap)
    pos += 1
    if (pos >= len) pos = 0
  }
}

class BAALLLightControl(configFile: String) extends ActWithStash with ActorLogging {
  var modemap: Map[String, LightMode] = Map()
  var clock: Long = 500
  var url = ""
  var rgbMap = Map("r" -> "1", "g" -> "2", "b" -> "3")
  var cmd: Option[LightMode] = None
  var currentMode = ""

  def parseVal(xml: Node): Seq[RGBVal] = {
    xml match {
      case <set></set> =>
        List(RGBVal(
        Integer.parseInt((xml \ "@red").text),
        Integer.parseInt((xml \ "@green").text),
        Integer.parseInt((xml \ "@blue").text),
        tryParseInt((xml \ "@id").text,"") match {
          case "" => None
          case s => Some(Integer.parseInt(s))
        }
      ))
      case _ => List()
    }
  }

  def parseMode(xml: Node): Seq[RGBVal] = {
    xml match {
      case <loop>{contents@_*}</loop> => contents.flatMap(parseVal)
      case _ => parseVal(xml)
    }
  }

  def tryParseInt(key: String, default: String) = {
    try {
      Integer.parseInt(key).toString
    } catch {
      case e: NumberFormatException => default
    }
  }

  def parseRGBMap(xml: Node): Map[String, String] = {
    val parse = (x:Node,key:String,default:String) =>{
      val temp = (xml \ key).text
      if (temp.isEmpty) default
      else temp
    }

    Map("r" -> parse(xml,"@redarg","1"),
        "g" -> parse(xml,"@greenarg","2"),
        "b" -> parse(xml, "@bluearg","3"))
  }

  override def preStart() {
    try {
      val rurl = this.getClass.getClassLoader.getResource(configFile)
      val config = if (rurl != null) {
        log.info("resource: " + rurl)
        scala.xml.XML.load(rurl)
      } else scala.xml.XML.loadFile(configFile)
      clock = Integer.parseInt((config \ "@clockspeed").text).toLong
      rgbMap = parseRGBMap(config)
      for (mode <- config \\ "mode") {
        val modename = (mode \ "@name").text
        modemap = modemap.updated(modename, new LightMode(mode.child.flatMap(parseMode)))
      }
      url = (config \ "@url").text
      log.info("url " + url)
    }
    catch {
      case e: FileNotFoundException => log.warning("no xml light config file: " + configFile)
      case e: SAXParseException => log.error(e, "xml error in light config file: " + configFile)
      case e: Throwable => log.error(e, "other xml error in light config file: " + configFile)
    }
    context.setReceiveTimeout(clock.millisecond)
  }

  override def receive = {
    case ReceiveTimeout => cmd match {
      case Some(mode) => mode.step(url, rgbMap)
      case None =>
    }
    case s: String => modemap.get(s) match {
      case v@Some(mode) =>
        mode.first = true
        mode.step(url, rgbMap)
        cmd = v
        currentMode = s
      case None => if (!url.isEmpty) log.info("unknown light mode: " + s)
    }
    case _ => log.error("cannot process cmd")
  }
}
