package src.BAALL

import xml.Elem
import xml.pull._

/**
 * Created with IntelliJ IDEA.
 * User: dodi
 * Date: 06.09.12
 * Time: 10:05
 * this class provides the functionality to read the next complete xml element of a xml streams and to convert the
 * obtained events into a scala xml element.
 */
object XmlRead {

  /**
   *
   * @param er am XMLEventReader
   * @return the next complete sequence of xmlelements, i.e., <tag>...</tag> or an empty list if nothing is available
   */
  def getNextEvents(key: String, er: XMLEventReader): List[XMLEvent] = {
    if (er.hasNext) {
      var counter = 1
      var item = er.next()
      while (!item.isInstanceOf[EvElemStart] && er.hasNext) {
        item = er.next()
      }
      var items = List[XMLEvent](item)
      while (counter > 0 && er.hasNext) {
        er.next() match {
          case EvElemStart(prefix, label, attr, scope) => counter = counter + 1; items = items ++ List[XMLEvent](EvElemStart(prefix, label, attr, scope))
          case EvElemEnd(prefix, label) => items = items ++ List[XMLEvent](EvElemEnd(prefix, label)); counter = counter - 1
          case EvText(text) => items = items ++ List[XMLEvent](EvText(text))
          case EvEntityRef(entity) => items = items ++ List[XMLEvent](EvEntityRef(entity))
          case x => println("skipping xml event " + x)
        }
      }
      if (counter > 0) Nil else items
    }
    else List[XMLEvent]()
  }

  /**
   * converts a list of xml elements to an xml elem
   * @param events as list
   * @return an xml element
   */
  def getXML(events: List[XMLEvent]): Seq[scala.xml.Node] = {
    events match {
      case EvElemStart(prefix, label, attr, scope) :: r =>
        var counter = 1
        var (childs, rest) = r.span(
          (p: XMLEvent) => {
            p match {
              case EvElemStart(_, _, _, _) => counter = counter + 1
              case EvElemEnd(_, _) => counter = counter - 1
              case _ =>
            }
            counter > 0
          }
        )
        val elem = Elem(prefix, label, attr, scope, true, getXML(childs): _*)
        //elem = elem.copy(child = elem.child ++ getXML(childs))
        assert(rest.head.isInstanceOf[EvElemEnd])
        Seq[scala.xml.Node](elem) ++ getXML(rest.drop(1))
      case EvText(txt) :: r => scala.xml.Text(txt) ++ getXML(r)
      case _ => Seq[scala.xml.Node]()
    }
  }
}
