package src.math

/**
 * Created with IntelliJ IDEA.
 * User: autexier
 * Date: 31.01.2013
 * Time: 11:17
 */

case class Point(x: Double, y: Double) {
  def distanceTo(other: Point): Double = StrictMath.hypot(x - other.x, y - other.y)
}

case class Triangle(A: Point, B: Point, C: Point) {
  val lengthAB: Double = A.distanceTo(B)
  val lengthAC: Double = A.distanceTo(C)
  val lengthBC: Double = C.distanceTo(B)

  val angleABC = StrictMath.acos(// beta
    (lengthAC * lengthAC - lengthBC * lengthBC - lengthAB * lengthAB) /
      (-2 * lengthBC * lengthAB)).toDegrees
  val angleBCA = StrictMath.acos(// gamma
    (lengthAB * lengthAB - lengthAC * lengthAC - lengthBC * lengthBC) /
      (-2 * lengthBC * lengthAC)).toDegrees
  val angleCAB = StrictMath.acos(// alpha
    (lengthBC * lengthBC - lengthAB * lengthAB - lengthAC * lengthAC) /
      (-2 * lengthAB * lengthAC)).toDegrees

  val heightA: Double = lengthAC * StrictMath.sin(angleBCA.toRadians)
  val heightB: Double = lengthAB * StrictMath.sin(angleCAB.toRadians)
  val heightC: Double = lengthBC * StrictMath.sin(angleABC.toRadians)
}

case class Segment(x: Point, y: Point) {
  def length: Double = x.distanceTo(y)

  def distanceTo(c: Point): Double = {
    val triangle = Triangle(x, y, c)
    if (triangle.angleABC > 90) {
      // degenerate case
      //println("Degenerate ABC")
      c.distanceTo(y)
    } else if (triangle.angleCAB > 90) {
      // degenerate case
      //println("Degenerate CAB")
      c.distanceTo(x)
    } else triangle.heightC
  }
}
