grammar Map;

options {
  output=AST;
}

@lexer::header {
package src.navi;
}

@parser::header {
package src.navi;

import java.awt.*;
}

@parser::members {
WPMap map = new WPMap();

public WPMap getMap()
{
  return map;
  }
}

WS  : (' ' |'\r' | '\t'|'\u000C'|'\n') {$channel=HIDDEN;}
    ;

LINE_COMMENT
    : '//' ~('\n'|'\r')* '\r'? '\n' {$channel=HIDDEN;}
    ;

MULTIPLE_LINE_COMMENT
    : '/**' (options {greedy=false;} : . )* '*/' {$channel=HIDDEN;}
    ;

ID  : ('-'|'a'..'z'|'A'..'Z'|'0'..'9'|'_')*
    ;

map : approxbezierpoints upperleftcorner upperrightcorner downleftcorner upperleftexternalcorner upperrightexternalcorner downleftexternalcorner block+ EOF
    ;

approxbezierpoints
    : nr=ID { map.setApproxbezier(Integer.parseInt($nr.getText())); }
    ;
upperleftcorner
    : x=ID y=ID { map.setUpperleftcorner(new Point(Integer.parseInt($x.getText()),Integer.parseInt($y.getText())));}
    ;
upperrightcorner
    : x=ID y=ID { map.setUpperrightcorner(new Point(Integer.parseInt($x.getText()),Integer.parseInt($y.getText())));}
    ;
downleftcorner
    : x=ID y=ID { map.setDownleftcorner(new Point (Integer.parseInt($x.getText()),Integer.parseInt($y.getText()))); }
    ;
upperleftexternalcorner
    : x=ID y=ID { map.setUpperleftcornerexternal(new Point (Integer.parseInt($x.getText()), Integer.parseInt($y.getText()))); }
    ;
upperrightexternalcorner
    : x=ID y=ID { map.setUpperrightcornerexternal(new Point (Integer.parseInt($x.getText()), Integer.parseInt($y.getText()))); }
    ;
downleftexternalcorner
    : x=ID y=ID { map.setDownleftcornerexternal(new Point (Integer.parseInt($x.getText()), Integer.parseInt($y.getText()))); }
    ;
block
    : name=ID x=ID y=ID count=ID edges[Integer.parseInt($count.text),$name.getText()] { map.addWayPoint($name.getText(), Integer.parseInt($x.getText()), Integer.parseInt($y.getText())); }
    ;

edges[int nr, String source]
    : {$nr==1}?=> edge[$source]
    | edge[$source] edges[$nr-1, $source]
    ;
edge[String source]
    : target=ID 'L' { map.addEdge($source,$target.getText(),"L"); }
    | target=ID 'O' { map.addEdge($source,$target.getText(),"O"); }
    ;
