package src.ship

import scala.collection.immutable.Nil
import scala.util.parsing.combinator.JavaTokenParsers

/**
 * Created with IntelliJ IDEA.
 * User: autexier
 * Date: 13.06.2012
 * Time: 17:04
 */

trait TBoxFormulaParser extends JavaTokenParsers with ConceptParser {
  val mypartition = (xs: List[Concept]) => {
    var i = 0
    var result = List.empty[List[Concept]]
    while (i < xs.length - 1) {
      result = result ++ List(xs.takeRight(xs.length - i))
      i += 1
    }
    result flatMap {
      case hd :: tl => tl map (p => Pair(hd, p))
      case _ => Nil
    }
  }

  def tboxformula: Parser[Ontology] = {
    "Disjoint" ~ "(" ~ rep1sep(concept, ",") ~ ")" ^^ {
      case _ ~ _ ~ cs ~ _ =>
        Ontology(mypartition(cs) map (p => ConceptEquality(NotConcept(UniversalConcept), AndConcept(p._1, p._2))),
          Nil, Nil)
    } |
      conceptADT |
      concept ~ "<" ~ rep1sep(concept, "|" | "<+>") ^^ {
        case c1 ~ "<" ~ cs => {
          cs match {
            case c :: xs =>
              Ontology(List(ConceptInclusion(c1, (c /: xs)((cl, cr) => OrConcept(cl, cr)))) ++ {
                mypartition(cs) map (p => ConceptEquality(NotConcept(UniversalConcept), AndConcept(p._1, p._2)))
              }, Nil, Nil)
            case Nil => EmptyOntology
          }
        }
      } | concept ~ "=" ~ rep1sep(concept, "|" | "<+>") ^^ {
      case c1 ~ "=" ~ cs => {
        cs match {
          case c :: xs => Ontology(List(ConceptEquality(c1, (c /: xs)((cl, cr) => OrConcept(cl, cr)))) ++ {
            mypartition(cs) map (p => ConceptEquality(NotConcept(UniversalConcept), AndConcept(p._1, p._2)))
          }, Nil, Nil)
          case Nil => EmptyOntology
        }
      }
    }
  }

  def conceptADT: Parser[Ontology] = {
    concept ~ "::=" ~ rep1sep(tboxconstructorconcept, "|") ^^ {
      case adt ~ _ ~ (constr :: constrs) => {
        val temp = Ontology(
          ConceptEquality(adt, (constrs :\ constr._1)((consdecl, concept) => OrConcept(consdecl._1, concept))) ::
            (mypartition(constr :: constrs map (x => x._1)) map
              (p => ConceptEquality(NotConcept(UniversalConcept), AndConcept(p._1, p._2))))
              ++ Nil,
          Nil, Nil)
        (temp /: (constr :: constrs))((ont, consdecl) => ont.extendBy(consdecl._2))
      }
      case _ => EmptyOntology
    }
  }

  def tboxconstructorconcept: Parser[(Concept, Ontology)] = {
    concept ~ opt("(" ~ repsep(role ~ ":" ~ concept, ",") ~ ")") ^^ {
      case c ~ x => x match {
        case None => (c, EmptyOntology)
        case Some(_ ~ selDecls ~ _) =>
          val newSelDecls = selDecls map {
            case NamedRole(r) ~ _ ~ (codom@DataConcept(_, _, _)) => (DataRole(r), codom)
            case r ~ _ ~ codom => (r, codom)
          }
          (c, Ontology(
          {
            val temp = newSelDecls map {
              case (r, codomain) =>
                AndConcept(MaxRestriction(r, 1, codomain), MinRestriction(r, 1, codomain))
            }
            temp match {
              case Nil => Nil
              case c1 :: cs => List(ConceptInclusion(c, (cs :\ c1)((x, y) => AndConcept(x, y))))
            }
          },
          (newSelDecls map {
            case (r@DataRole(_), codomain) => List(RoleDeclaration(r, c, codomain))
            case (r, codomain) =>
              List(RoleDeclaration(r, c, codomain), RoleFun(r)) ++ {
                if (codomain == c) List(RoleIrr(r)) else Nil
              }
          }).flatten, Nil))
      }
    }
  }
}

abstract class OntologyAxiom {
  def getConcepts: Set[ConceptConstant]

  def getRoles: Set[NamedRole]

  def getIndividuals: Set[String]
}

/* TBox Axioms */
sealed abstract class TBoxFormula extends OntologyAxiom

case class ConceptInclusion(sub: Concept, sup: Concept) extends TBoxFormula {
  override def toString = sub + " < " + sup

  override def getConcepts: Set[ConceptConstant] = sub.getConcepts ++ sup.getConcepts

  override def getRoles: Set[NamedRole] = sub.getRoles ++ sup.getRoles

  override def getIndividuals: Set[String] = sub.getIndividuals ++ sup.getIndividuals

}

case class ConceptEquality(left: Concept, right: Concept) extends TBoxFormula {
  override def toString = (left, right) match {
    case (NotConcept(UniversalConcept), AndConcept(c1, c2)) =>
      "Disjoint(" + c1 + ", " + c2 + ")"
    case _ => left + " = " + right
  }

  override def getConcepts: Set[ConceptConstant] = left.getConcepts ++ right.getConcepts

  override def getRoles: Set[NamedRole] = left.getRoles ++ right.getRoles

  override def getIndividuals: Set[String] = left.getIndividuals ++ right.getIndividuals
}

trait RBoxFormulaParser extends JavaTokenParsers with ConceptParser {
  /* RBOX FORMULA SYNTAX:
   RBOXFORMULA     ::= ROLEDECLARATION  | ROLEINCLUSION | ROLEEQUALITY | ROLEPROPERTY
   ROLEDECLARATION ::= ROLE ":" CONCEPT "*" CONCEPT
   ROLEINCLUSION   ::= ROLE ("." ... "." ROLE)? < ROLE
   ROLEEQUALITY    ::= ROLE "=" ROLE ("." ... "." ROLE)?
   ROLEPROPERTY    ::= "Sym(" ROLE ")" | "Trans(" ROLE ")" | "Ref(" ROLE ")" |
                       "Irref(" ROLE ")" | "Dis(" ROLE ")" | "Func(" ROLE ")" |
                       "Asym(" ROLE ")"
  */
  def rboxformulas: Parser[List[RBoxFormula]] =
    role ~ ":" ~ concept ~ ("*" | "->" | "<->" | "<-") ~ (concept | dataConcept) ^^ {
      case r ~ _ ~ domain ~ s ~ range =>
        val rr = (r, range) match {
          case (NamedRole(nr), DataConcept(_, _, _)) =>
            DataRole(nr)
          case _ => r
        }
        val rd = RoleDeclaration(rr, domain, range)
        s match {
          case "<->" => List(rd, RoleFun(rr), RoleInvFun(rr))
          case "->" => List(rd, RoleFun(rr))
          case "<-" => List(rd, RoleInvFun(rr))
          case "*" => List(rd)
        }
    } |
      repsep(role, ".") ~ "<" ~ role ^^ {
        case rs ~ "<" ~ r2 => List(RoleInclusion(rs, r2))
      } | role ~ "=" ~ repsep(role, ".") ^^ {
      case r1 ~ "=" ~ rs => List(RoleEquality(rs, r1))
    } | ("Sym" | "Trans" | "Ref" | "Irref" | "Func" | "Invfunc" | "Asym") ~ "(" ~ role ~ ")" ^^ {
      case s ~ _ ~ r ~ _ => s match {
        case "Sym" => List(RoleSym(r))
        case "Trans" => List(RoleTra(r))
        case "Ref" => List(RoleRef(r))
        case "Irref" => List(RoleIrr(r))
        case "Func" => List(RoleFun(r))
        case "Invfunc" => List(RoleInvFun(r))
        case "Asym" => List(RoleAsym(r))
      }
    } | "Dis" ~ "(" ~ role ~ "," ~ role ~ ")" ^^ {
      case "Dis" ~ _ ~ r1 ~ _ ~ r2 ~ _ => List(RoleDis(r1, r2))
    }
}


/* RBox Axioms */
sealed abstract class RBoxFormula extends OntologyAxiom

/* 0. RBox Declaration */
case class RoleDeclaration(role: Role, domain: Concept, range: Concept) extends RBoxFormula {
  override def toString = role + ":" + domain + " * " + range

  override def getConcepts: Set[ConceptConstant] = role.getConcepts ++ domain.getConcepts ++ range.getConcepts

  override def getRoles: Set[NamedRole] = role.getRoles ++ domain.getRoles ++ range.getRoles

  override def getIndividuals: Set[String] = role.getIndividuals ++ domain.getIndividuals ++ range.getIndividuals
}

/* 1. RBox Hierarchy */
case class RoleInclusion(rolecomposition: List[Role], role: Role) extends RBoxFormula {
  override def toString = rolecomposition.mkString(".") + " < " + role

  override def getConcepts: Set[ConceptConstant] = ((role :: rolecomposition) :\ Set.empty[ConceptConstant])((r, s) => s ++ r.getConcepts)

  override def getRoles: Set[NamedRole] = ((role :: rolecomposition) :\ Set.empty[NamedRole])((r, s) => s ++ r.getRoles)

  override def getIndividuals: Set[String] = ((role :: rolecomposition) :\ Set.empty[String])((r, s) => s ++ r.getIndividuals)
}

case class RoleEquality(rolecomposition: List[Role], right: Role) extends RBoxFormula {
  override def toString = right + " = " + rolecomposition.mkString(".")

  override def getConcepts: Set[ConceptConstant] = ((right :: rolecomposition) :\ Set.empty[ConceptConstant])((r, s) => s ++ r.getConcepts)

  override def getRoles: Set[NamedRole] = ((right :: rolecomposition) :\ Set.empty[NamedRole])((r, s) => s ++ r.getRoles)

  override def getIndividuals: Set[String] = ((right :: rolecomposition) :\ Set.empty[String])((r, s) => s ++ r.getIndividuals)
}

/* 2. RBox Assertions */
case class RoleSym(role: Role) extends RBoxFormula {
  override def toString = "Sym(" + role + ")"

  override def getConcepts = Set.empty[ConceptConstant]

  override def getRoles = role.getRoles

  override def getIndividuals = role.getIndividuals
}

case class RoleTra(role: Role) extends RBoxFormula {
  override def toString = "Trans(" + role + ")"

  override def getConcepts = Set.empty[ConceptConstant]

  override def getRoles = role.getRoles

  override def getIndividuals = role.getIndividuals
}

case class RoleRef(role: Role) extends RBoxFormula {
  override def toString = "Ref(" + role + ")"

  override def getConcepts = Set.empty[ConceptConstant]

  override def getRoles = role.getRoles

  override def getIndividuals = role.getIndividuals
}

case class RoleIrr(role: Role) extends RBoxFormula {
  override def toString = "Irref(" + role + ")"

  override def getConcepts = Set.empty[ConceptConstant]

  override def getRoles = role.getRoles

  override def getIndividuals = role.getIndividuals
}

case class RoleDis(role1: Role, role2: Role) extends RBoxFormula {
  override def toString = "Dis(" + role1 + ", " + role2 + ")"

  override def getConcepts = Set.empty[ConceptConstant]

  override def getRoles = role1.getRoles ++ role2.getRoles

  override def getIndividuals = role1.getIndividuals ++ role2.getIndividuals
}

case class RoleFun(role: Role) extends RBoxFormula {
  override def toString = "Func(" + role + ")"

  override def getConcepts = Set.empty[ConceptConstant]

  override def getRoles = role.getRoles

  override def getIndividuals = role.getIndividuals
}

case class RoleInvFun(role: Role) extends RBoxFormula {
  override def toString = "Invfunc(" + role + ")"

  override def getConcepts = Set.empty[ConceptConstant]

  override def getRoles = role.getRoles

  override def getIndividuals = role.getIndividuals
}

case class RoleAsym(role: Role) extends RBoxFormula {
  override def toString = "Asym(" + role + ")"

  override def getConcepts = Set.empty[ConceptConstant]

  override def getRoles = role.getRoles

  override def getIndividuals = role.getIndividuals
}

sealed trait VarOrConstant extends ObjectLevelVariable {
  def name: String

  def matches(v: VarOrConstant): Option[Substitution]

  def getConsts: Set[String] = Set.empty

  override def toString = name
}

case class Var(name: String) extends VarOrConstant {
  var isExplicitVar = false

  def matches(v: VarOrConstant) = {
    val s = Some(/(v, this))
    v match {
      case Var(x) => if (x == name) Some(Id()) else s
      case _ => s
    }
  }

  override def freeVars = Set(this)
}

case class Const(name: String) extends VarOrConstant {

  def matches(v: VarOrConstant) = {
    v match {
      case Const(x) => if (x == name) Some(Id()) else None
      case _ => None
    }
  }

  override def getConsts = Set(name)
}

case class Literal(name: String) extends VarOrConstant {
  // "name" is bad for datatype literals but I don't want to change it globally

  def matches(v: VarOrConstant) = {
    v match {
      case Literal(x) => if (x == name) Some(Id()) else None
      case _ => None
    }
  }
}

trait ABoxFormulaParser extends JavaTokenParsers with ConceptParser {
  def aboxformulas: Parser[List[ABoxFormula]] = repsep(aboxformula, ",")

  def aboxformula: Parser[ABoxFormula] = conceptassertion | roleassertion |
    differentfromassertion | sameasassertion

  def differentfromassertion: Parser[ABoxFormula] = varOrConst ~ "!=" ~ varOrConst ^^ {
    case var1 ~ "!=" ~ var2 => DifferentFromAssertion(var1, var2)
  }

  def sameasassertion: Parser[ABoxFormula] = varOrConst ~ "==" ~ varOrConst ^^ {
    case var1 ~ "==" ~ var2 => SameAsAssertion(var1, var2)
  }

  def conceptassertion: Parser[ABoxFormula] = varOrConst ~ ":" ~ concept ^^ {
    case s ~ ":" ~ c => ConceptAssertion(s, c)
  }

  def roleassertion: Parser[ABoxFormula] = "(" ~ varOrConst ~ "," ~ varOrConstOrLiteral ~ ")" ~ ":" ~ role ^^ {
    case "(" ~ n1 ~ "," ~ n2 ~ ")" ~ ":" ~ r => RoleAssertion(n1, n2, n2 match {
      case Literal(_) => r match {
        case NamedRole(n) => DataRole(n)
        case _ => {
          println("complex role '" + r + "' used as datatype!")
          r
        }
      }
      case _ => r
    })
  }

  def varOrConst: Parser[VarOrConstant] =
    ":" ~ string ^^ {
      case _ ~ s => Const(s)
    } |
      "?" ~ string ^^ {
        case _ ~ s => new Var(s) {
          isExplicitVar = true
        }
      } |
      string ^^ {
        case s => Var(s)
      }

  def varOrConstOrLiteral: Parser[VarOrConstant] = literal ^^ {
    case s => Literal(s)
  } | varOrConst

  /**
   * parses an abox formula and unwraps the result
   */
  def parseABox(s: String): ABoxFormula = {
    parseAll(aboxformula, s) match {
      case Success(res, in) => res
      case _ => throw new Exception("cannot parse formula " + s)
    }
  }

  /**
   * parses a list of abox formulas separated by ,
   */
  def parseABoxs(s: String): List[ABoxFormula] = {
    parseAll(aboxformulas, s) match {
      case Success(res, in) => res
      case _ => throw new Exception("cannot parse formulas " + s)
    }
  }
}

/* ABox Atoms are a:C, (a,B):R, (a,b): not R, a neq b */
sealed abstract class ABoxFormula extends OntologyAxiom with ObjectLevelVariable with GroundableObject[ABoxFormula] with JavaLogging {
  def matches(o: ABoxFormula): Option[Substitution]

  def groundBy(exceptions: Set[ObjectLevelVariable]): ABoxFormula = {
    var gsubst: Substitution = Id()
    this.freeVars filterNot exceptions.contains map {
      case rv@Var(s) if !rv.isExplicitVar =>
        // log.info(v + " is no explicit var")
        gsubst = CompSubst(/(Const(s), Var(s)), gsubst)
      case _ => Id()
    }
    gsubst(this)
  }

  /*
    Checks if the formula is a concept assertion or role assertion with literal concepts/roles.
   */
  def isPrimitive: Boolean = this match {
    case ConceptAssertion(n, c) => c.isLiteral
    case RoleAssertion(s, t, r) => r.isLiteral
    case _ => false
  }
}

case class ConceptAssertion(name: VarOrConstant, concept: Concept) extends ABoxFormula {
  override def toString = name + ":" + concept

  def matches(f: ABoxFormula) = f match {
    case ConceptAssertion(n, c) => if (c == concept) name.matches(n) else None
    case _ => None
  }

  override def freeVars: Set[ObjectLevelVariable] = (name match {
    case v@Var(x) => Set[ObjectLevelVariable](v)
    case _ => Set.empty[ObjectLevelVariable]
  }) ++ concept.freeVars

  override def getConcepts: Set[ConceptConstant] = concept.getConcepts

  override def getRoles: Set[NamedRole] = concept.getRoles

  override def getIndividuals: Set[String] = name.getConsts ++ concept.getIndividuals
}

case class DifferentFromAssertion(var1: VarOrConstant, var2: VarOrConstant) extends ABoxFormula {
  override def toString = var1 + " != " + var2

  override def matches(f: ABoxFormula) = None

  override def freeVars = var1.freeVars ++ var2.freeVars

  override def getConcepts: Set[ConceptConstant] = Set.empty

  override def getRoles: Set[NamedRole] = Set.empty

  override def getIndividuals: Set[String] = (Set(var1, var2) :\ Set.empty[String])((name, s) => s ++ name.getConsts)
}

case class SameAsAssertion(var1: VarOrConstant, var2: VarOrConstant) extends ABoxFormula {
  override def toString = var1 + " == " + var2

  override def matches(f: ABoxFormula) = None

  override def freeVars = var1.freeVars ++ var2.freeVars

  override def getConcepts: Set[ConceptConstant] = Set.empty

  override def getRoles: Set[NamedRole] = Set.empty

  override def getIndividuals: Set[String] = (Set(var1, var2) :\ Set.empty[String])((name, s) => s ++ name.getConsts)
}

case class RoleAssertion(src: VarOrConstant, trg: VarOrConstant, role: Role) extends ABoxFormula {
  override def toString = "(" + src + ", " + trg + "):" + role

  def matches(f: ABoxFormula): Option[Substitution] = {
    val res = f match {
      case RoleAssertion(s, t, r) => if (r == role) {
        src.matches(s) match {
          case Some(s1) => s1(trg).matches(t) match {
            case Some(s2) => Some(CompSubst(s2, s1))
            case _ => None
          }
          case _ => None
        }
      } else None
      case _ => None
    }
    //println("Match "+this+" to "+f+" = "+res)
    res
  }

  override def freeVars: Set[ObjectLevelVariable] = src.freeVars ++ trg.freeVars

  override def getConcepts: Set[ConceptConstant] = role.getConcepts

  override def getRoles: Set[NamedRole] = role.getRoles

  override def getIndividuals: Set[String] = role.getIndividuals ++ (Set(src, trg) :\ Set.empty[String])((name, s) =>
    s ++ name.getConsts)
}

case class Disjunction(f: ABoxFormula, g: ABoxFormula) extends ABoxFormula {
  override def toString = f + " or " + g

  def matches(phi: ABoxFormula): Option[Substitution] = {
    phi match {
      case Disjunction(f1, g1) => f.matches(f1) match {
        case Some(s) => s.apply(g).matches(g1)
        case _ => None
      }
      case _ => None
    }
  }

  override def freeVars: Set[ObjectLevelVariable] = f.freeVars ++ g.freeVars

  override def getConcepts: Set[ConceptConstant] = f.getConcepts ++ g.getConcepts

  override def getRoles: Set[NamedRole] = f.getRoles ++ g.getRoles

  override def getIndividuals: Set[String] = f.getIndividuals ++ g.getIndividuals
}
