package src.ship

/**
 * Created with IntelliJ IDEA.
 * User: autexier
 * Date: 18.07.2012
 * Time: 13:16
 */

trait ProgramAndMonitorParser extends ABoxFormulaParser {
  protected override val whiteSpace = """(\s|//.*|(?m)/\*(\*(?!/)|[^*])*\*/)+""".r

  /* BNF
     DLDECLARATION = action A opt((parameters)) = {
                            pre = List[AboxFormula]
                            opt(effect = List[AboxFormula])
                            star(if (List[AboxFormula]) opt(then) NonEmptyList[AboxFormula])
                            } |
                     function f opt((parameters)) = { DLPROG  }  |
                     indirect effect N = {
                     indirect effect N = {
                       init =  List[AboxFormula]
                       causes = List[AboxFormula]
                       cond = List[AboxFormula]
                     }
                     */

  def dldecl: Parser[Map[String, DLDeclaration]] = rep(declaction | declfunction | declindirecteffect) ^^ {
    case xs => (xs :\ Map.empty[String, DLDeclaration])((decl, m) => decl match {
      case Action(n, _, _, _, _, _) => m updated(n, decl)
      case Function(n, _, _) => m updated(n, decl)
      case IndirectEffect(n, _, _, _) => m updated(n, decl)
    })
  }

  def declaction: Parser[Action] = "action" ~ string ~ parameterdecls ~ "=" ~ "{" ~ actionpre ~ opt(uncondeffect) ~ condeffects ~ opt(actionexec) ~ "}" ^^ {
    case _ ~ name ~ parameterdecls ~ _ ~ _ ~ pre ~ opteffect ~ condeffects ~ Some(Some(execeffects)) ~ _ =>
      val effect = opteffect match {
        case Some(e) => e
        case None => UnconditionalEffect(Nil, Nil)
      }
      Action(name, parameterdecls, pre, Some(execeffects), effect, condeffects)
    case _ ~ name ~ parameterdecls ~ _ ~ _ ~ pre ~ opteffect ~ condeffects ~ _ ~ _ =>
      val effect = opteffect match {
        case Some(e) => e
        case None => UnconditionalEffect(Nil, Nil)
      }
      Action(name, parameterdecls, pre, None, effect, condeffects)
  }

  def actionexec: Parser[Option[Pair[java.lang.reflect.Method, List[Var]]]] = {
    "exec" ~ "=" ~ string ~ "." ~ string ~ parameterdecls ^^ {
      case _ ~ _ ~ classname ~ _ ~ methodname ~ pars => {
        var c: Option[java.lang.Class[_]] = scala.None
        var m: Option[java.lang.reflect.Method] = scala.None
        try {
          c = Some(Class.forName("src.ship.external." + classname))
          m = Some(selectActionMethod(c.get, methodname, pars.length))
          if (m.get.getReturnType != Class.forName("scala.Option")) {
            println("Method " + methodname + " has invalid return type " + m.get.getReturnType)
            m = None
          }
        }
        catch {
          case e: ClassNotFoundException => println("Unknown class " + "src.ship.external." + classname)
          case e: NoSuchMethodException => println("Unknown method " + methodname + " for class " + c.get)
          case e: SecurityException => println("Cannot used method " + methodname + " for security reasons")
        }
        m match {
          case Some(method) => Some(Pair(method, pars))
          case _ => None
        }
      }
    }
  }

  private def selectActionMethod(c: java.lang.Class[_], methodname: String, argnumber: Int): java.lang.reflect.Method = {
    val ontclass = Class.forName("src.ship.Ontology")
    val stringclass = Class.forName("java.lang.String")
    argnumber match {
      case 0 => c.getMethod(methodname, ontclass)
      case 1 => c.getMethod(methodname, ontclass, stringclass)
      case 2 => c.getMethod(methodname, ontclass, stringclass, stringclass)
      case 3 => c.getMethod(methodname, ontclass, stringclass, stringclass, stringclass)
      case 4 => c.getMethod(methodname, ontclass, stringclass, stringclass, stringclass, stringclass)
      case 5 => c.getMethod(methodname, ontclass, stringclass, stringclass, stringclass, stringclass, stringclass)
      case 6 => c.getMethod(methodname, ontclass, stringclass, stringclass, stringclass, stringclass, stringclass, stringclass)
      case 7 => c.getMethod(methodname, ontclass, stringclass, stringclass, stringclass, stringclass, stringclass, stringclass, stringclass)
      case 8 => c.getMethod(methodname, ontclass, stringclass, stringclass, stringclass, stringclass, stringclass, stringclass, stringclass, stringclass)
      case 9 => c.getMethod(methodname, ontclass, stringclass, stringclass, stringclass, stringclass, stringclass, stringclass, stringclass, stringclass, stringclass)
      case 10 => c.getMethod(methodname, ontclass, stringclass, stringclass, stringclass, stringclass, stringclass, stringclass, stringclass, stringclass, stringclass, stringclass)
      case _ => throw new NoSuchMethodException()
    }
  }

  def declindirecteffect: Parser[IndirectEffect] = "indirect" ~ "effect" ~ string ~ "=" ~ "{" ~ indeffectinit ~ indeffectcauses ~ indeffectcond ~ "}" ^^ {
    case _ ~ _ ~ name ~ _ ~ _ ~ init ~ causes ~ cond ~ _ =>
      IndirectEffect(name, init, causes, cond)
  }

  def aboxformulasordeleteind: Parser[Pair[List[ABoxFormula], List[VarOrConstant]]] = repsep(aboxformulaprim | deleteindividuals, ",") ^^ {
    case forms => {
      var formulas = List.empty[ABoxFormula]
      var individuals = List.empty[VarOrConstant]
      for (x <- forms) {
        x match {
          case abf: ABoxFormula => formulas = abf :: formulas
          case _ => individuals = individuals ++ x.asInstanceOf[List[VarOrConstant]]
        }
      }
      Pair(formulas, individuals)
    }
  }

  def deleteindividuals: Parser[List[VarOrConstant]] = "delete" ~ "(" ~ repsep(string, ",") ~ ")" ^^ {
    case _ ~ _ ~ stringlist ~ _ => stringlist map ((s: String) => Var(s))
  }

  def parameterdecls: Parser[List[Var]] = opt("(" ~ repsep(string, ",") ~ ")") ^^ {
    case Some(_ ~ stringlist ~ _) => stringlist map ((s: String) => Var(s))
    case _ => Nil
  }

  def actionpre: Parser[List[ABoxFormula]] = "pre" ~ "=" ~ repsep(aboxformulaprim, ",") ^^ {
    case _ ~ _ ~ e => e
  }

  def indeffectinit: Parser[List[ABoxFormula]] = "init" ~ "=" ~ repsep(aboxformulaprim, ",") ^^ {
    case _ ~ _ ~ e => e
  }

  def indeffectcond: Parser[List[ABoxFormula]] = "cond" ~ "=" ~ repsep(aboxformulaprim, ",") ^^ {
    case _ ~ _ ~ e => e
  }

  def indeffectcauses: Parser[UnconditionalEffect] = "causes" ~ "=" ~ aboxformulasordeleteind ^^ {
    case _ ~ _ ~ e => UnconditionalEffect(e._1, e._2)
  }

  def uncondeffect: Parser[UnconditionalEffect] = "effect" ~ "=" ~ aboxformulasordeleteind ^^ {
    case _ ~ _ ~ e => UnconditionalEffect(e._1, e._2)
  }

  def condeffects: Parser[ConditionalEffect] = rep(condeffect) ^^ {
    case ms => ConditionalEffect((Map.empty[List[ABoxFormula], UnconditionalEffect] /: ms)((m1, m2) => m1 ++ m2))
  }

  def condeffect: Parser[Map[List[ABoxFormula], UnconditionalEffect]] = "if" ~ "(" ~ repsep(aboxformulaprim, ",") ~ ")" ~ opt("then") ~ aboxformulasordeleteind ^^ {
    case _ ~ _ ~ conditions ~ _ ~ _ ~ effects => Map[List[ABoxFormula], UnconditionalEffect] {
      conditions -> UnconditionalEffect(effects._1, effects._2)
    }
  }

  def declfunction: Parser[Function] = ("process" | "function") ~ string ~ parameterdecls ~ "=" ~ "{" ~ dlprog ~ "}" ^^ {
    case _ ~ name ~ parameterdecls ~ _ ~ _ ~ body ~ _ =>
      Function(name, parameterdecls, body.groundBy(parameterdecls.toSet))
  }

  /*
     DLPROG ::= DLPAR
     DLPAR  ::= DLALT OPT(| DLPAR)
     DLALT ::= DLSEQ OPT((+|+>) DLALT)
     DLSEQ ::= DLPOSTFIX OPT(; DLSEQ)
     DLPOSTFIX ::= ABOXAND ? | DLPROGPRAM OPT(*)
     DLPROGPRAM ::= if ( ABOXOR ) then DLPROGPRAM else | while (ABOXOR) ( DLPROG ) |
                    switch ( case ABOXAND => DLPROG )+

     ABOXOR ::= ABOXAND  or ABOXTEST
     ABOXAND ::= ABOXPRIM and ABOXAND
     ABOXPRIM ::= "not" ABOXFORMULA | ABOXFORMULA
   */

  def dlprog: Parser[DLProgram] = dlalt ~ opt("|" ~ dlprog) ^^ {
    case p1 ~ Some(_ ~ p2) => Par(p1, p2)
    case p1 ~ None => p1
  }

  def dlalt: Parser[DLProgram] = dlseq ~ opt(("+" | "+>") ~ dlalt) ^^ {
    case p1 ~ Some(_ ~ p2) => Alt(p1, p2)
    case p1 ~ None => p1
  }

  def dlseq: Parser[DLProgram] = dlpostfix ~ opt(";" ~ dlseq) ^^ {
    case p1 ~ Some(_ ~ p2) => Seq(p1, p2)
    case p1 ~ None => p1
  }

  def dlpostfix: Parser[DLProgram] =
    aboxand ~ "?" ^^ {
      case (phi ~ _) => Test(phi)
    } |
      dlprim ~ opt("*") ^^ {
        case p ~ Some(_) => Star(p)
        case p ~ None => p
      } |
      string ~ opt("(" ~ repsep(string, ",") ~ ")") ~ opt("*") ^^ {
        case name ~ pars ~ star => {
          val body = pars match {
            case Some(_ ~ parameters ~ _) => Appl(name, parameters map (x => Var(x)))
            case None => Appl(name, Nil)
          }
          star match {
            case Some(_) => Star(body)
            case None => body
          }
        }
      }

  def dlprim: Parser[DLProgram] = "(" ~ dlprog ~ ")" ^^ {
    case _ ~ p ~ _ => p
  } |
    "{" ~ dlprog ~ "}" ^^ {
      case _ ~ p ~ _ => p
    } |
    "if" ~ "(" ~ aboxand ~ ")" ~ dlprog ~ "else" ~ dlprog ^^ {
      case _ ~ _ ~ phi ~ _ ~ p1 ~ _ ~ p2 => Alt(Seq(Test(phi), p1), Seq(TestNot(phi), p2))
    } |
    "while" ~ "(" ~ aboxand ~ ")" ~ dlprog ^^ {
      case _ ~ _ ~ phi ~ _ ~ p =>
        // Star(Alt(Seq(Test(phi),p),TestNot(phi)))
        Seq(Star(Switch(List[Guard](Guard(phi, p)))), TestNot(phi))
    } |
    "switch" ~ rep("case" ~ (aboxand | "_") ~ "=>" ~ dlprog) ^^ {
      case _ ~ cases => Switch(cases map {
        case "case" ~ phis ~ "=>" ~ prg => if (phis.isInstanceOf[String]) Guard(List.empty[ABoxFormula], prg)
        else Guard(phis.asInstanceOf[List[ABoxFormula]], prg)
      })
    } |
    "forall" ~ aboxand ~ "=>" ~ dlprog ^^ {
      case _ ~ phis ~ _ ~ prg => Forall(phis, prg)
    } |
    "init" ~ quotedstring ~ foltl ^^ {
      case _ ~ desc ~ phi => MonitorExec(phi, desc, phi)
    } | "init" ~ string ~ parameterdecls ^^ {
    case _ ~ name ~ pars => Pred(name, pars)
  } | "skip" ^^ {
    case _ => Skip
  }

  def aboxand: Parser[List[ABoxFormula]] = aboxformulaprim ~ opt("and" ~ aboxand) ^^ {
    case phi ~ Some(_ ~ psis) => phi :: psis
    case phi ~ None => List(phi)
  }

  def aboxformulaprim: Parser[ABoxFormula] = "not" ~ aboxformula ^^ {
    case _ ~ phi => OntologyFun.Neg(phi)
  } |
    aboxformula

  def foltl: Parser[FOLTL[Concept, ABoxFormula]] = foltlor ~ opt("=>" ~ foltl) ^^ {
    case phi ~ Some("=>" ~ psi) => FOLTL.Impl(phi, psi)
    case phi ~ None => phi
  }

  def foltlprefix: Parser[FOLTL[Concept, ABoxFormula]] = "(" ~ foltl ~ ")" ^^ {
    case "(" ~ f ~ ")" => f
  } |
    folquant | folunary | folconstant | folatom


  def foltlor: Parser[FOLTL[Concept, ABoxFormula]] = foltland ~ opt("or" ~ foltlor) ^^ {
    case phi ~ Some("or" ~ psi) => FOLTL.Or(phi, psi)
    case phi ~ None => phi
  }

  def foltland: Parser[FOLTL[Concept, ABoxFormula]] = foltluntil ~ opt("and" ~ foltland) ^^ {
    case phi ~ Some("and" ~ psi) => And(phi, psi)
    case phi ~ None => phi
  }

  def foltluntil: Parser[FOLTL[Concept, ABoxFormula]] = foltlprefix ~ opt("U" ~ foltlprefix) ^^ {
    case phi ~ Some("U" ~ psi) => U(phi, psi)
    case phi ~ None => phi
  }


  def folconstant: Parser[FOLTL[Concept, ABoxFormula]] = ("true" | "false") ^^ {
    case "true" => True
    case "false" => FOLTL.False
  }

  def folatom: Parser[FOLTL[Concept, ABoxFormula]] =
    aboxformula ^^ {
      case phi => Atom(phi)
    } |
      string ~ parameterdecls ^^ {
        case name ~ actpars => DefPred(Pred(name, actpars))
      }


  def folunary: Parser[FOLTL[Concept, ABoxFormula]] = ("not" | "F" | "G" | "X") ~ foltl ^^ {
    case "not" ~ phi => Not(phi)
    case "F" ~ phi => FOLTL.F(phi)
    case "G" ~ phi => FOLTL.G(phi)
    case "X" ~ phi => X(phi)
  } | "run" ~ dlprog ^^ {
    case _ ~ p => RunProg(p)
  }

  def folquant: Parser[FOLTL[Concept, ABoxFormula]] = ("all" | "ex") ~ repsep(folvardecl, ",") ~ "." ~ foltl ^^ {
    case quant ~ maps ~ "." ~ phi =>
      val decls: Map[String, Concept] = (maps :\ Map.empty[String, Concept])((m1, m2) => m1 ++ m2)
      (decls.keySet :\ phi)((s, psi) => if (quant.equalsIgnoreCase("all")) All(Var(s), decls.get(s).get, psi)
      else FOLTL.Ex(Var(s), decls.get(s).get, psi))
  }

  def folvardecl: Parser[Map[String, Concept]] = repsep(string, ",") ~ ":" ~ concept ^^ {
    case rs ~ ":" ~ c => (Map.empty[String, Concept] /: rs)((m, n) => m updated(n, c))
  }

}
