package src.ship

import akka.actor.ActorDSL._
import akka.actor.{Props, ActorSystem, ActorRef}
import java.awt.{GraphicsDevice, GraphicsEnvironment, Font, Color}
import java.io.{FileReader, File}
import java.lang.reflect.Constructor
import scala.swing._
import scala.swing.event.{EditDone, CaretUpdate, ButtonClicked}
import src.BAALL._
import src.ship.rest.BAALLApp
import src.simple.SimpleGUI

/**
 * Created with IntelliJ IDEA.
 * User: autexier
 * Date: 3.09.2012
 * Time: 09:32
 */

abstract class SHIPEngineGUICommand

case class Run(p: DLProgram) extends SHIPEngineGUICommand

case class Stop() extends SHIPEngineGUICommand

case class Show(s: String) extends SHIPEngineGUICommand

case class StartClient(c: String) extends SHIPEngineGUICommand

case class StopClient(c: String) extends SHIPEngineGUICommand

case object StopClients extends SHIPEngineGUICommand

case class Help() extends SHIPEngineGUICommand

case class SaveOnt(file: String) extends SHIPEngineGUICommand

case class UCHPolling(on: Boolean) extends SHIPEngineGUICommand

case class MonitorDecl(name: String, pars: List[Var], desc: String, phi: FOLTL[Concept, ABoxFormula]) extends SHIPEngineGUICommand

case class Monitor(phi: FOLTL[Concept, ABoxFormula], desc: String) extends SHIPEngineGUICommand

case class MonitorPred(name: String, pars: List[Const]) extends SHIPEngineGUICommand

case class QueryCmd(phis: List[ABoxFormula]) extends SHIPEngineGUICommand

case class Update(phis: List[ABoxFormula], deleteIndividuals: List[Const]) extends SHIPEngineGUICommand

case class ResetOntology(o: Ontology) extends SHIPEngineGUICommand

case class ResetProcesses(pl: List[String]) extends SHIPEngineGUICommand

case class ResetMonitors(ml: List[String]) extends SHIPEngineGUICommand

case object ResetSHIPTool extends SHIPEngineGUICommand


trait SHIPToolParser extends ProgramAndMonitorParser with OntologyParser {

  def ercmd: Parser[Any] =
    "trace" ^^ {
      case _ => Trace
    } |
    "pause" ~ decimalNumber ^^ {
      case _ ~ s => try {
        val i = Integer.parseInt(s)
        log.warning("Waiting "+i)
        Thread.sleep(i)
        Trace
      } catch { case e:NumberFormatException => wait(100);Trace }
    } |
      "startclient" ~ string ^^ {
        case _ ~ s => StartClient(s)
      } |
      "stopclients" ^^ {
        case _ => StopClients
      } |
      "stopclient" ~ string ^^ {
        case _ ~ s => StopClient(s)
      } |
      saveOnt |
      "ontology" ^^ {
        case _ => Show("#ONTOLOGY")
      } |
      "show" ~ string ^^ {
        case "show" ~ n => Show(n)
      } |
      "help" ^^ {
        case "help" => Help()
      } |
      "stop" ^^ {
        case "stop" => Stop()
      } |
      "baallsync" ~ ("on" | "off") ^^ {
        case _ ~ "on" => UCHPolling(on = true)
        case _ ~ "off" => UCHPolling(on = false)
      } |
      dlquery | dlrunprog | runmonitor | dladdmonitor | dlaboxupdate | dlresetparser |
      loadfile | dldecl

  def saveOnt: Parser[Any] = "saveont" ~ string ^^ {
    case _ ~ filename => SaveOnt(filename)
  }

  def dlquery: Parser[Any] = "query" ~ repsep(aboxformulaprim, ",") ^^ {
    case _ ~ formulas =>
      QueryCmd(formulas)
  }

  def dlrunprog: Parser[Any] = "run" ~ dlpostfix ^^ {
    case _ ~ p => Run(p.ground)
  }

  def runmonitor: Parser[Any] = "init" ~ string ~ parameterdecls ^^ {
    case _ ~ name ~ pars =>
      val pred = Pred(name, pars map {
        case Var(n) => Const(n)
      })
      Monitor(DefPred(pred), pred.toString())
  } | "init" ~ quotedstring ~ foltl ^^ {
    case _ ~ desc ~ phi => Monitor(phi.ground, desc)
  }

  def dladdmonitor: Parser[MonitorDecl] = "monitor" ~ string ~ parameterdecls ~ "=" ~ quotedstring ~ foltl ^^ {
    case _ ~ name ~ pars ~ _ ~ desc ~ phi => MonitorDecl(name, pars, desc, phi)
  }

  def dlaboxupdate: Parser[Any] = "update" ~ aboxformulasordeleteind ^^ {
    case _ ~ part => Update(part._1 map (_.ground), (part._2 map {
      case Var(s) => Const(s)
      case v => v
    }).asInstanceOf[List[Const]])
  }

  def dlresetparser: Parser[List[SHIPEngineGUICommand]] = "reset" ~ opt(repsep("abox" | "proc" | "mon", ",")) ^^ {
    case "reset" ~ x => x match {
      case Some(a :: xs) => ((a :: xs) :\ List.empty[SHIPEngineGUICommand])((s, c) => s match {
        case "abox" => ResetOntology(Ontology(List.empty[TBoxFormula], List.empty[RBoxFormula], List.empty[ABoxFormula])) :: c
        case "proc" => ResetProcesses(Nil) :: c
        case "mon" => ResetMonitors(Nil) :: c
      })
      case _ => List(ResetSHIPTool)
    }
  }

  def cmdListParser: Parser[List[Any]] = repsep(ercmd, ";")

  def loadfile: Parser[Any] = "load" ~ pathparser ^^ {
    case "load" ~ filename =>
      log.info("loading file: " + filename)
      val file = new File(filename)
      val isOwlFile = filename.endsWith(".owl") || filename.endsWith(".rdf") || filename.endsWith("XML")
      var pa: Parser[Any] = cmdListParser
      if (file.exists() && file.isFile && file.canRead || isOwlFile) {
        val dotpos = filename lastIndexOf '.'
        // println("file " + filename + " has dot position:" + dotpos)
        if (dotpos >= 0) {
          //println(filename.substring(dotpos))
          pa = filename.substring(dotpos) match {
            case ".mon" => rep(dladdmonitor)
            case ".prg" => dldecl
            case ".ont" => ontology
            case _ => pa
          }
        }
        if (isOwlFile)
          (if (filename.startsWith("http://") || filename.startsWith("file://")) OWL2Ontology.fromIRIString(filename)
          else OWL2Ontology.fromFile(file)) match {
            case Some(o) => o
            case None =>
              log.error("could not load OWL ontology " + filename)
              Nil
          }
        else {
          //println("Parser = "+parser)
          parseAll(pa, new FileReader(file)) match {
            case Success(s, in) =>
              //println("Parse success " + s)
              s
            case NoSuccess(s, in) =>
              log.error("Parse error at " + in.pos + ": " + s)
              Nil
          }
        }
      } else {
        log.error("File not found or unreadable " + filename + "\n" +
          "CWD " + System.getProperty("user.dir"))
        Nil
      }
  }

  def processParseCmd(res: Any) {
    res match {
      case decls: Map[_, _] =>
        decls.values map SHIPTool.ref.!
      case value: List[_] =>
        value map processParseCmd
      case o: Ontology =>
        SHIPTool.ref ! ResetOntology(o)
      case _ =>
        SHIPTool.ref ! res
    }
  }

}

class DemoHelperDialog(font: swing.Font, titlestr: String, steps: List[(String, String)])
  extends Frame with SHIPToolParser {
  title = titlestr
  val buttons = steps map {
    case (menuText, c) =>
      val button = new Button(menuText) {
        font_=(font)
      }
      reactions += {
        case ButtonClicked(`button`) =>
          parseAll(ercmd, c) match {
            case Success(res, _) => processParseCmd(res)
            case NoSuccess(msg, _) => log.error("parse-error in '" + c + "' " + msg)
          }
      }
      button
  }
  contents = new GridPanel(buttons.length, 1) {
    contents ++= buttons
  }
  listenTo(buttons: _*)
}

case class SHIPMonitorStatus(phi: FOLTL[Concept, ABoxFormula],
                             desc: String,
                             lastvalid: FOLTL[Concept, ABoxFormula],
                             currentstate: FOLTL[Concept, ABoxFormula],
                             currentstatevalue: FOLTL[Concept, ABoxFormula],
                             futurestate: FOLTL[Concept, ABoxFormula],
                             futurestatevalue: FOLTL[Concept, ABoxFormula]) {
  override def toString: String = {
    phi match {
      case Not(True) => "" + phi + "\nlastvalid = " + lastvalid + "\ncurrentstate " + currentstate + " = " + currentstatevalue
      case _ => phi.toString
    }
  }
}

case class Version(major: Int, minor: Int) {
  override def toString: String = major.toString + "." + minor
}

object SHIPTool extends SimpleSwingApplication with JavaLogging {

  val version = Version(1, 0)

  var yamamotohostname = "localhost"
  //"baall-server.informatik.uni-bremen.de"
  var yamamotoport = 10280
  //10280

  val akkaSystem = ActorSystem("ship-akka-actors")
  val ref = akkaSystem.actorOf(Props[SHIPTool], "ship")
  val er = akkaSystem.actorOf(Props(new SHIPEngineRoom(Map.empty[String, DLDeclaration])), "room")

  var fontsize = 12

  val labelfont = new swing.Font(null, Font.BOLD, fontsize)
  val textfont = new swing.Font(null, Font.PLAIN, fontsize)

  val yamamotoCons = classOf[Visualizer].getConstructor(classOf[String], classOf[Int], classOf[Boolean], classOf[ActorRef])
  val rollandCons = classOf[RollandClient].getConstructors()(0)

  var declaredClientTools: Map[String, Triple[Option[ActorRef], Constructor[_], Array[AnyRef]]] =
    Map("Yamamoto" -> ((None, yamamotoCons, Array("localhost", 10280: Integer, er))),
      "Yamamoto2" -> ((None, yamamotoCons, Array("localhost", 10281: Integer, er))),
      "RealBaall" -> ((None, classOf[Visualizer].getConstructor(classOf[String], classOf[Int], classOf[Boolean], classOf[ActorRef]),
        Array("baall-server.informatik.uni-bremen.de", 1028: Integer, false: java.lang.Boolean, er))),
      "BAALLApp" -> ((None, classOf[BAALLApp].getConstructor(classOf[Int]), Array(8000: Integer))), // warning: er not declared at this point
      "EmergencyButton" -> ((None, classOf[EmergencyButtonActor].getConstructor(), Array())),
      "BAALLLight" -> ((None, classOf[OntUpdaterBAALLLight].getConstructor(), Array())),
      "Rolland2" -> ((None, rollandCons, Array("192.168.178.99", 12349: Integer, "Rolland2", Set(SHIPTool.ref)))),
      "Rolland" -> ((None, rollandCons, Array("192.168.178.100", 12348: Integer, "Rolland", Set(SHIPTool.ref)))),
      "Simple" -> ((None, classOf[SimpleGUI].getConstructor(classOf[swing.Font], classOf[String]), Array(textfont, "Simple")))
    )

  var startupargs: Array[String] = Array.empty

  override def startup(args: Array[String]) {
    log.info("startup started")

    startupargs = args
    super.startup(args)
    args map {
      case "--baall=ship-labor" => yamamotohostname = "ship-server.cart"
      case "--baall=ship-labor-stahl" => yamamotohostname = "192.168.114.168"
      case "--baall=ship-labor-maeder" => yamamotohostname = "134.102.218.219"
      case "--baall" => yamamotohostname = "baall-server.informatik.uni-bremen.de"; yamamotoport = 1028
      case "--enable-webapp" => ref ! StartClient("BAALLApp")
      case arg =>
        val baallArg = "--baall="
        if (arg.startsWith(baallArg)) yamamotohostname = arg.substring(baallArg.length)
        val loadArg = "--load="
        if (arg.startsWith(loadArg)) ref ! ("load " + arg.substring(loadArg.length))
        val linknxArg="--linknx="
        if (arg.startsWith(linknxArg)) {
          val linknxhostport = arg.substring(linknxArg.length)
          val Pattern = "([^:]+):([0-9]+)".r
          linknxhostport match {
            case Pattern(host,port) =>
              yamamotohostname = host;yamamotoport = port.toInt
              log.info("Linknx on host %s at port %s".format(host,port))
            case _ =>
          }
        }
    }
    log.info("yamamotohost = " + yamamotohostname)
    log.info("yamamotoport = " + yamamotoport)
    declaredClientTools = declaredClientTools.updated("Yamamoto", (None, yamamotoCons, Array(yamamotohostname, yamamotoport: Integer, true:java.lang.Boolean, er)))
    declaredClientTools = declaredClientTools.updated("Yamamoto2", (None, yamamotoCons, Array(yamamotohostname, yamamotoport + 1: Integer, true:java.lang.Boolean, er)))
  }

  val processesPane = new ProcessesPane(textfont)
  val shipguiprocesses = akkaSystem.actorOf(Props(new SHIPGUIProcesses(processesPane)), "procs")
  val monitorsPane = new MonitorsPane(textfont)
  val shipguimonitors = akkaSystem.actorOf(Props(new SHIPGUIMonitors(monitorsPane)), "mons")
  val logPane = new LogPane(textfont)
  val shiproomlog = akkaSystem.actorOf(Props(new SHIPRoomLog(logPane)), "log")

  val status = new GridPanel(0, 2) {
    contents += processesPane
    contents += monitorsPane
  }

  var mainframe: MainFrame = null

  /*def quitAction: Action = Action("Quit") {
    Dialog.showConfirmation(mainPanel, "Are you sure you want to quit?",
      "Quit Game", Dialog.Options.YesNo) match {
        case Dialog.Result.Yes => window.dispose()
        case _ => None
     }
  }*/

  def setLocation(m: MainFrame) {
    var screen: Option[Int] = None
    startupargs map {
      arg => {
        val screenArg = "--screen="
        if (arg.startsWith(screenArg)) {
          screen = Some(StrictMath.max(1, arg.substring(screenArg.length).toInt))
        }
      }
    }
    val ge: GraphicsEnvironment = GraphicsEnvironment.getLocalGraphicsEnvironment
    val devices = ge.getScreenDevices

    val gd: GraphicsDevice = screen match {
      case None => ge.getDefaultScreenDevice
      case Some(i) =>
        if (i <= devices.length) devices(i - 1)
        else ge.getDefaultScreenDevice
    }

    log.debug("Screen " + screen + " device " + gd.getDefaultConfiguration.getBounds)
    val bounds = gd.getDefaultConfiguration.getBounds

    mainframe.peer.setLocation(
      (bounds.getCenterX - mainframe.peer.getSize.getWidth / 2.0).toInt,
      (bounds.getCenterY - mainframe.peer.getSize.getHeight / 2.0).toInt)
  }

  def top = {
    mainframe = new MainFrame with SHIPToolParser {
      title = "SHIPTool Version " + version
      menuBar = new MenuBar() {
        val m1 = new Menu(" Components ") {
          font = labelfont
          for (clientname <- declaredClientTools.keys) {
            val client = declaredClientTools.get(clientname).get._1
            contents += new Menu(clientname) {
              font = textfont
              contents += new MenuItem(scala.swing.Action("Start") {
                SHIPTool.ref ! StartClient(clientname)
              }) {
                font = textfont
              }
              contents += new MenuItem(scala.swing.Action("Stop") {
                SHIPTool.ref ! StopClient(clientname)
              }) {
                font = textfont
              }
              // Quick hack to add a menu for visualizer; should be done abstractly
              if (client.isDefined) {
                contents += new MenuItem(scala.swing.Action("Enable UCH Polling") {
                  client.get ! UCHPolling(on = true)
                }) {
                  font = textfont
                }
                contents += new MenuItem(scala.swing.Action("Disable UCH Polling") {
                  client.get ! UCHPolling(on = false)
                }) {
                  font = textfont
                }
              }
            }
          }
        }
        val m2 = new Menu("SHIPTool ") {
          font = labelfont
          /* contents += new MenuItem(scala.swing.Action("Zoom +") {
              SHIPTool.adjustFontSize(+5,SHIPTool.top)
            })
          contents += new MenuItem(scala.swing.Action("Zoom -") {
            SHIPTool.adjustFontSize(-5,SHIPTool.top)
          }) */
          contents += new MenuItem(scala.swing.Action("Reset") {
            SHIPTool.ref ! ResetSHIPTool
          }) {
            font = textfont
          }

          contents += new MenuItem(scala.swing.Action("Exit") {
            SHIPTool.ref ! Stop()
          }) {
            font = textfont
          }
        }
        var usecases: List[(String, DemoHelperDialog)] = List()
        val exampledir = new File("examples")
        if (exampledir.exists() && exampledir.isDirectory && exampledir.canRead) {
          for (file <- exampledir.listFiles) {
            val fileName = file.getName
            val suf = ".dem"
            if (fileName.endsWith(suf)) {
              val demoname = Builtin.upCase1(fileName.substring(0, fileName.length - suf.length))
              val demo = new DemoHelperDialog(textfont, demoname, DemoParser.getMenuLines(file))
              usecases = (demoname, demo) :: usecases
            }
          }
        }
        val m3 = new Menu("UseCases ") {
          font = labelfont
          for (usecase <- usecases) {
            contents += new MenuItem(scala.swing.Action(usecase._1) {
              SHIPTool.ref ! ResetSHIPTool
              usecase._2.visible = true
            })
          }
        }
        contents += m2
        contents += m1
        contents += m3
      }

      object cmdinput extends TextField(70) {
        font = textfont
        editable = true
        hasFocus
      }

      contents = new BoxPanel(Orientation.Vertical) {
        contents += status
        contents += new ScrollPane {
          viewportView = logPane
        }
        contents += new Label("Log") {
          font = labelfont
        }
        contents += cmdinput
        contents += new Label("Command") {
          font = labelfont
        }
        border = Swing.EmptyBorder(15, 10, 10, 10)
      }
      listenTo(cmdinput)
      peer.setLocationRelativeTo(null)
      peer.setAlwaysOnTop(false)
      reactions += {
        case CaretUpdate(`cmdinput`) => cmdinput.background_=(Color.WHITE)
        case EditDone(`cmdinput`) =>
          log.info("[" + cmdinput.text + "]")
          parseAll(ercmd, cmdinput.text) match {
            case NoSuccess(res, in) =>
              log.error("Command line parse failure at " + in.pos + ": " + res)
              cmdinput.background_=(new java.awt.Color(240, 125, 125))
            case Success(res, in) =>
              cmdinput.background_=(Color.WHITE)
              log.info("CMD Sending " + (res match {
                case o: Ontology => "\n" + o.toStringAux
                case _ => res.toString
              }))
              processParseCmd(res)
              cmdinput.text_=("")
          }
        /* case WindowActivated(_) =>
        if (false) {
          peer.setLocationRelativeTo(null)
          peer.setAlwaysOnTop(false)
          init = false
        }
        */
        case _ =>
      }
      preferredSize = new Dimension(400, 300)
    }
    setLocation(mainframe)
    mainframe
  }

  def adjustFontSize(offset: Int, x: UIElement) {
    log.info(":Adjustfontsize " + offset)
    val computefont = (f: swing.Font) => {
      val res = f.getSize + offset
      log.info("New size = " + res)
      if (res < 10) f.deriveFont(10)
      else if (res > 40) f.deriveFont(40)
      else f.deriveFont(res)
    }
    x match {
      case c: Container =>
        c.contents map (y => adjustFontSize(offset, y))
        c match {
          case rw: RichWindow => rw.menuBar.menus map (y => adjustFontSize(offset, y))
          case m: Menu => m.font_=(computefont(m.font))
          case _ =>
        }
      case _ =>
    }
    x.peer.setIgnoreRepaint(false)
    //x.font_=(computefont(x.font))
    x.peer.setFont(computefont(x.peer.getFont))
    x.peer.repaint()
  }

  for (actor <- List[ActorRef](SHIPTool.shiproomlog, SHIPTool.shipguimonitors, SHIPTool.shipguiprocesses))
    SHIPTool.er ! AddObserver(actor)
  ref ! ResetSHIPTool
  ref ! StartClient("EmergencyButton")
}

class SHIPTool extends ActWithStash with SHIPToolParser {
  var monitorDecls = Map.empty[String, MonitorDecl]

  override def receive = {
    case ResetSHIPTool =>
      SHIPTool.er ! ResetProcesses(Nil)
      SHIPTool.er ! ResetMonitors(Nil)
      SHIPTool.er ! ResetOntology(EmptyOntology)
    case StartClient(s) => SHIPTool.declaredClientTools get s match {
      case Some((client, cons, args)) =>
        SHIPTool.logPane.log("Starting client " + s)
        client match {
          case Some(a) => SHIPTool.er ! AddObserver(a)
          case None =>
            val a = context.actorOf(Props(cons.newInstance(args: _*).asInstanceOf[ActWithStash]), s)
            SHIPTool.declaredClientTools = SHIPTool.declaredClientTools.updated(s, (Some(a), cons, args))
            SHIPTool.er ! AddObserver(a)
        }
      case None => SHIPTool.logPane.log("Unknown client tool " + s)
    }
    case StopClient(s) => SHIPTool.declaredClientTools get s match {
      case Some((Some(client), _, _)) =>
        SHIPTool.logPane.log("Stopping client " + s)
        // todo: actors are not stopped, but run through!
        client ! src.ship.Stop()
        SHIPTool.er ! RemoveObserver(client)
      case Some((None, _, _)) =>
      case None => SHIPTool.logPane.log("Unknown client tool " + s)
    }
    case StopClients => SHIPTool.declaredClientTools.map(v => {
      v._2 match {
        case (Some(a), _, _) =>
          a ! src.ship.Stop()
          SHIPTool.er ! RemoveObserver(a)
          SHIPTool.logPane.log("Stopping client " + v._1)
        case _ =>
      }
    })
    case v@UCHPolling(_) =>
      log.info("shiptool receiving " + v)
      SHIPTool.declaredClientTools.values.map(
        a => {
          if (a._1.isDefined)
            a._1.get ! v
        })
    case v@Trace => SHIPTool.er ! v
    case Show(n) =>
      if (monitorDecls.contains(n)) {
        val temp: MonitorDecl = monitorDecls.get(n).get
        SHIPTool.logPane.log("" + temp.name + "(" + temp.pars + ") = " + temp.desc + " " + temp.phi)
      } else SHIPTool.er ! Show(n)
    case Stop() =>
      SHIPTool.quit()
    case Help() =>
      val s = "* Help * \n" +
        "======== \n" +
        "load path-to-filename.ext : loads the file (ext=prg process functions and action,\n" +
        "                            ext=mon monitoring formulas, ext=ont for ontologies)\n" +
        "stop  : stop and quit the SHIPTool\n" +
        "trace : chronological list of executed actions, monitors and updates obtained from the outside world\n" +
        "help  : this help message\n" +
        "(startclient|stopclient) N : starts resp. stops the client process N\n" +
        "reset (proc|mon|abox)* : stops all running processes, monitors and ontology\n" +
        "                        without argument resets all\n" +
        "ontology : pretty prints the current ontology\n" +
        "show s   : pretty prints the declaration of monitor, action or function named 's'\n" +
        "action A(X1, ... Xn) = ABODY : declares the action A with arguments X1, ... Xn\n" +
        "                              ABODY   ::= '{ pre =' (ABOX| N'!=' N)(',' (ABOX| (N'!='N)))* \n" +
        "                                            'effect = '(ABOX|DELETE)(, (ABOX|DELETE))* \n" +
        "                                            ('if ('ABOX(and ABOX)*')' (ABOX|DELETE)(',' (ABOX|DELETE))*)*\n" +
        "                                            'exec =' SCALAOBJECTCLASSNAME'.'METHODNAME('('N, ..., N')')? }\n" +
        "                              DELETE  ::= 'delete (' N(, N)* ')'\n" +
        "                              ABOX    ::= 'not'? (N':'CONCEPT | '('N','N'):'ROLE)\n" +
        "                              CONCEPT ::= C | '~' CONCEPT | CONCEPT ('&'|'+') CONCEPT\n" +
        "                                        | ('all'|'ex') ROLE '.' CONCEPT | '{' N '}' | '@'N'.'CONCEPT\n" +
        "                              ROLE    ::= R | '~' ROLE | ROLE ('&'|'+') ROLE\n" +
        "                                        | '{('N','N')}' \n" +
        "function F(X1, ... Xn) = { FBODY } : declares the function F with arguments X1 ... Xn\n" +
        "                         FBODY ::= A'('N (',' N)*')' | F'('N(',' N)*')'\n" +
        "                                 | FBODY '|' FBODY | FBODY* \n" +
        "                                 | 'switch' ('case' ABOX(',' ABOX)* '=>' FBODY)*\n" +
        "                                 | 'while' '(' ABOX(',' ABOX)* ')' FBODY\n" +
        "                                 | FBODY '+' FBODY | '<'ABOX(,and ABOX)*'?>'\n" +
        "                                 | 'init' P'('N(',' N)*')' | 'init' FOLTL\n" +
        "indirect effect N = INDEFFECTBODY : declares the indirect effect rule N\n" +
        "                    INDEFFECTBODY ::= '{ init =' ABOX(',' ABOX)* \n" +
        "                                        'causes = '(ABOX|DELETE)(, (ABOX|DELETE))* \n" +
        "                                        'cond = ' ABOX(, ABOX)* }\n" +
        "monitor P(x1, ... xn) = FOLTL : declares the monitor P with arugments X1, ... Xn\n" +
        "                      FOLTL ::= ABOX | FOLTL ('and'|'or'|'=>'|'U') FOLTL\n" +
        "                              | ('not'|'X'|'F'|'G') FOLTL | '(' FOLTL ')' \n" +
        "                              | ('all'|'ex') N':'CONCEPT . FOLTL\n" +
        "                              | P'('N(',' N)*')' | 'run' FBODY \n" +
        "init P(n1, .. nk)   : starts the monitor P with names n1...\n" +
        "init FOLTL          : starts FOLTL as a monitor\n" +
        "run F(n1, .. nk)    : starts the process (action or function) with names n1...nk\n" +
        "update (ABOX|DELETE)(, (ABOX|DELETE))* : updates the ABOX\n" +
        "query ABOX(, ABOX)* : queries the current ontology for the conjunction of the ABOX formulas\n" +
        "saveont filename    : saves the current ontology as an OWL ontology in the designated file\n"
      SHIPTool.logPane.log(s)
    case Run(p) => SHIPTool.er ! StartProgram(p)
    case Monitor(phi, desc) => SHIPTool.er ! StartMonitor(phi, desc)
    case v@MonitorDecl(n, pars, desc, phi) =>
      monitorDecls = monitorDecls updated(n, v)
      SHIPTool.er ! AddDeclaration(n, Predicate(n, pars, desc, phi))
    case MonitorPred(n, pars) =>
      if (monitorDecls contains n) {
        val mdecl = (monitorDecls get n).get
        mdecl match {
          case MonitorDecl(name, formpars, desc, phi1) =>
            if (formpars.length == pars.length) {
              val subst: Substitution =
                ((0 until pars.length) :\ (Id(): Substitution))((i, s) => {
                  log.info(i + " " + s)
                  CompSubst(/(pars(i), Var(formpars(i).name)), s)
                })
              //println("Subst "+subst)
              val phi: FOLTL[Concept, ABoxFormula] = subst.applyF(phi1)
              if (phi.freeVars.isEmpty) SHIPTool.er ! StartMonitor(phi, subst.applyS(desc))
              else log.error("Monitor formula contains free variables " + phi)
            }
        }
      }
    case f@Function(n, _, _) => SHIPTool.er ! AddDeclaration(n, f)
    case a@Action(n, _, _, _, _, _) => SHIPTool.er ! AddDeclaration(n, a)
    case v@IndirectEffect(n, _, _, _) => SHIPTool.er ! AddDeclaration(n, v)
    case Update(formulas, deleteIndividuals) => SHIPTool.er ! UpdateAbox(formulas, deleteIndividuals)
    case QueryCmd(formulas) => SHIPTool.er ! QueryOnt(formulas)
    case SaveOnt(filename) => SHIPTool.er ! SaveOntology(filename)
    case ResetOntology(abox) => SHIPTool.er ! ResetOntology(abox)
    case ResetMonitors(ml) => SHIPTool.er ! ResetMonitors(ml)
    case ResetProcesses(pl) => SHIPTool.er ! ResetProcesses(pl)
    case s: String => parseAll(ercmd, s) match {
      case Success(result, _) => processParseCmd(result)
      case _ => //
    }
    case x: RollandClientMessage => SHIPTool.er ! x
    case _ => //println("SHIPTool main: Ignoring Message " + c)
  }
}
