package src.ship

/**
 * Created with IntelliJ IDEA.
 * User: autexier
 * Date: 13.06.2012
 * Time: 17:18
 */

sealed abstract class Substitution {
  var map: Map[String, VarOrConstant]

  override def toString = map.toString()

  def apply(a: Var): VarOrConstant = {
    map.get(a.name) match {
      case Some(n) => n
      case None => a
    }
  }

  def apply(a: VarOrConstant): VarOrConstant = a match {
    case Var(n) => apply(Var(n))
    case _ => a
  }

  def apply(a: ABoxFormula): ABoxFormula = a match {
    case ConceptAssertion(e, c) => ConceptAssertion(apply(e), c)
    case RoleAssertion(e1, e2, r) => RoleAssertion(apply(e1), apply(e2), r)
    case DifferentFromAssertion(v1, v2) => DifferentFromAssertion(apply(v1), apply(v2))
    case SameAsAssertion(v1, v2) => SameAsAssertion(apply(v1), apply(v2))
    case Disjunction(v1, v2) => Disjunction(apply(v1), apply(v2))
  }

  def apply(al: List[ABoxFormula]): List[ABoxFormula] =
    al map (a => this.apply(a))

  def applyF(f: FOLTL[Concept, ABoxFormula]): FOLTL[Concept, ABoxFormula] = f match {
    case Atom(a) => Atom(apply(a))
    case DefPred(Pred(n, actpars)) => DefPred(Pred(n, actpars map (v => apply(v))))
    case RunProg(prog) => RunProg(prog.subst(this))
    case True => True
    case Not(g) => Not(applyF(g))
    case And(g, h) => And(applyF(g), applyF(h))
    case All(v, t, h) => All(v, applyC(t), |(this, v.asInstanceOf[Var]).applyF(h))
    case U(phi1, phi2) => U(applyF(phi1), applyF(phi2))
    case X(phi) => X(applyF(phi))
  }

  def applyS(s: String): String = {
    var result = s
    val it = map.keysIterator
    while (it.hasNext) {
      val key = it.next()
      val value = map.getOrElse(key, Var(key)) match {
        case Var(x) => "%" + x + "%"
        case Const(x) => x
        case Literal(x) => x
      }
      result = result.replace("%" + key + "%", value)
    }
    //println("Subst apply "+this+" on "+s+" = "+result)
    result
  }

  def applyC(c: Concept): Concept = {
    c match {
      case AndConcept(c1, c2) => AndConcept(applyC(c1), applyC(c2))
      case OrConcept(c1, c2) => OrConcept(applyC(c1), applyC(c2))
      case ForallRole(r, sub) => ForallRole(applyR(r), applyC(sub))
      case ExistsRole(r, sub) => ExistsRole(applyR(r), applyC(sub))
      case MaxRestriction(r, i, sub) => MaxRestriction(applyR(r), i, applyC(sub))
      case MinRestriction(r, i, sub) => MinRestriction(applyR(r), i, applyC(sub))
      case NotConcept(sub) => NotConcept(applyC(sub))
      case Nominal(v) => Nominal(apply(v))
      case _ => c
    }
  }

  def applyR(r: Role): Role = {
    r match {
      case AndRole(c1, c2) => AndRole(applyR(c1), applyR(c2))
      case OrRole(c1, c2) => OrRole(applyR(c1), applyR(c2))
      case NotRole(sub) => NotRole(applyR(sub))
      case _ => r
    }
  }
}

case class Id() extends Substitution {
  var map = Map.empty[String, VarOrConstant]
}

case class CompSubst(f: Substitution, g: Substitution) extends Substitution {
  var map = (g.map mapValues (v => f(v))) ++ (f.map -- g.map.keys)
}

case class |(s: Substitution, v: Var) extends Substitution {
  var map = s.map - v.name
}

case class /(c: VarOrConstant, n: Var) extends Substitution {
  var map = Map[String, VarOrConstant](n.name -> c)
}
