package src.ship.external

import src.ship._


/**
 * Created by IntelliJ IDEA.
 * User: dodi
 * Date: 7/27/12
 * Time: 3:00 PM
 * This class contains functions to convert from scala datatypes to ontology objects
 */

object Scala2Ontology {

  /**
   * converts a list the internal ontology format based on roles for tails, head and Concepts for NonEmtpy/Empty lists
   * @param l: list to be converted
   * @param pred: predecessor element of the list or ""
   * @param tailRole role representing the tail relation
   * @param headRole role representing the head relation
   * @param emptyClass class representing empty list elemts
   * @param nonEmptyClass class representing the nonempty list elements
   *
   */

  def list2Abox(l:List[String], pred:String, tailRole:String, headRole:String,
                   emptyClass:String, nonEmptyClass:String, ont:Ontology) : (List[ABoxFormula],String) =
  {
    //println("list2abox "+l)
    val c= ont.newIndividualGenerator
    l match {
      case x::xs =>
        (list2Abox(xs, c, tailRole, headRole, emptyClass, nonEmptyClass, ont)._1 ++
        (if (pred.isEmpty) List() else List(RoleAssertion(Const(pred), Const(c), NamedRole(tailRole)))) ++
          List[ABoxFormula](
            //ConceptAssertion(Const(c), ConceptConstant(nonEmptyClass)),
            RoleAssertion(Const(c), Const(x), NamedRole(headRole))
          ),c)
      case List() =>
        ( ConceptAssertion(Const(c), ConceptConstant(emptyClass)) ::
          (if (pred.isEmpty) List() else List(RoleAssertion(Const(pred), Const(c), NamedRole(tailRole)))),c)
    }
  }

  def pair2Abox(p:Pair[String,String],pairclass:String,fstrole:String,sndrole:String,ont:Ontology):(List[ABoxFormula],String) = {
    val pind = ont.newIndividualGenerator
    p match {
      case (x1,x2) => (List[ABoxFormula](
                    //ConceptAssertion(Const(pind),ConceptConstant(pairclass)),
                    RoleAssertion(Const(pind),Const(x1),NamedRole(fstrole)),
                    RoleAssertion(Const(pind),Const(x2),NamedRole(sndrole))
                    ), pind)
    }
  }
}