package src.test;

import com.clarkparsia.pellet.owlapiv3.PelletReasoner;
import org.semanticweb.owlapi.apibinding.OWLManager;
import org.semanticweb.owlapi.model.*;
import org.semanticweb.owlapi.reasoner.Node;
import org.semanticweb.owlapi.reasoner.NodeSet;

import java.util.Set;

/**
 * Created by IntelliJ IDEA.
 * User: dodi
 * Date: 7/17/12
 * Time: 12:04 PM
 * To change this template use File | Settings | File Templates.
 */
public class InstanceTest {
    public static void main(String[] args) throws OWLOntologyCreationException {
        String ont = "http://www.mindswap.org/2004/owl/mindswappers#";

		// create an ontology manager
		OWLOntologyManager manager = OWLManager.createOWLOntologyManager();
		OWLDataFactory factory = manager.getOWLDataFactory();

		// read the ontology
		OWLOntology ontology = manager.loadOntology( IRI.create(ont) );

		// load the ontology to the reasoner
		PelletReasoner reasoner = com.clarkparsia.pellet.owlapiv3.PelletReasonerFactory.getInstance().createReasoner( ontology );

		// create property and resources to query the reasoner
		OWLClass Person = factory.getOWLClass(IRI.create("http://xmlns.com/foaf/0.1/Person"));
		OWLObjectProperty workHomepage = factory.getOWLObjectProperty(IRI.create("http://xmlns.com/foaf/0.1/workInfoHomepage"));
		OWLDataProperty foafName = factory.getOWLDataProperty(IRI.create("http://xmlns.com/foaf/0.1/name"));

		// get all instances of Person class
		Set<OWLNamedIndividual> individuals = reasoner.getInstances( Person, false ).getFlattened();
		for(OWLNamedIndividual ind : individuals) {
		    // get the info about this specific individual
			Set<OWLLiteral> names = reasoner.getDataPropertyValues( ind, foafName );
		    NodeSet<OWLClass> types = reasoner.getTypes( ind, true );
		    NodeSet<OWLNamedIndividual> homepages = reasoner.getObjectPropertyValues( ind, workHomepage );

		    // we know there is a single name for each person so we can get that value directly
		    String name = names.iterator().next().getLiteral();
			System.out.println( "Name: " + name );

			// at least one direct type is guaranteed to exist for each individual
		    OWLClass type = types.iterator().next().getRepresentativeElement();
			System.out.println( "Type:" + type );

		    // there may be zero or more homepages so check first if there are any found
			if( homepages.isEmpty() ) {
				System.out.print( "Homepage: Unknown" );
			}
			else {
				System.out.print( "Homepage:" );
				for( Node<OWLNamedIndividual> homepage : homepages ) {
					System.out.print( " " + homepage.getRepresentativeElement().getIRI() );
				}
			}
		    System.out.println();
		    System.out.println();
		}
    }
}
