package src.test;

import aterm.AFun;
import aterm.ATerm;
import aterm.ATermAppl;
import com.clarkparsia.owlapiv3.OWL;
import com.clarkparsia.pellet.owlapiv3.PelletReasoner;
import com.clarkparsia.pellet.owlapiv3.PelletReasonerFactory;
import org.mindswap.pellet.KnowledgeBase;
import org.mindswap.pellet.Role;
import org.mindswap.pellet.utils.ATermUtils;
import org.semanticweb.owlapi.model.*;
import org.semanticweb.owlapi.reasoner.Node;
import org.semanticweb.owlapi.reasoner.NodeSet;
import org.semanticweb.owlapi.reasoner.impl.OWLNamedIndividualNode;
import src.ship.Builtin;

/**
 * Created by IntelliJ IDEA.
 * User: dodi
 * Date: 7/10/12
 * Time: 10:23 AM
 * This ontologyURI demonstrates the use of the pellet reasoner using the OWL API to load the data
 */
public class OWLAPITest {
    private static final String ontologyURI = "file:examples/transport.owl";

    public static void main(String[] args) throws OWLOntologyCreationException {
        // Create an OWLAPI manager that allows to load an ontology ontologyURI and
        // create OWLEntities
        OWLOntologyManager manager = OWL.manager;
        OWLOntology ontology = manager.loadOntology(IRI.create(ontologyURI));

        PelletReasoner reasoner = PelletReasonerFactory.getInstance().createReasoner(ontology);

        // we add a new individuum at update the at relation
        ATermAppl newind = ATermUtils.makeTermAppl(Builtin.namespace() + "rolland3");
        reasoner.getKB().addIndividual(newind);
        reasoner.getKB().addType(newind, ATermUtils.makeTermAppl(Builtin.namespace() + "WheelChair"));
        reasoner.getKB().addPropertyValue(ATermUtils.makeTermAppl(Builtin.namespace() + "at"),
                ATermUtils.makeTermAppl(Builtin.namespace() + "rolland3"),
                ATermUtils.makeTermAppl(Builtin.namespace() + "kitchenliving"));

        // we try to add a new equivalent class with the existing roles...
        //ATermAppl sw = ATermUtils.makeTermAppl(Builtin.namespace() + "NewClass");
        //reasoner.getKB().addEquivalentClass(ATermUtils.makeTermAppl(Builtin.namespace() + "MyClass"),
        //       ATermUtils.makeTermAppl(Builtin.namespace() + "WheelChair"));
        //        ATermUtils.makeSomeValues(ATermUtils.makeTermAppl(Builtin.namespace() + "at"),
        //        ATermUtils.makeTermAppl(Builtin.namespace() + "Free")));

        reasoner.getKB().realize();

        reasoner.getKB().printClassTree();

        printABox(reasoner);
        printRBox(reasoner);
        printTBox(reasoner);
        printClasses(reasoner);

        OWLDataFactory factory = manager.getOWLDataFactory();

        // create class to query the reasoner
        //OWLClass position = factory.getOWLClass(IRI.create(ontologyURI + "#Position"));
        OWLClass position = factory.getOWLClass(IRI.create(Builtin.namespace() + "Position"));
        System.out.println("Position: " + position);
        // get all instances of Person class
        System.out.println("Positions: " + reasoner.getInstances(position, false));
        System.out.println("SomeWhere: " + reasoner.getInstances(factory.getOWLClass(IRI.create(Builtin.namespace() + "SomeWhere")), false));
        System.out.println("NewClass: " + reasoner.getInstances(factory.getOWLClass(IRI.create(Builtin.namespace() + "NewClass")), false));

        // this is for getting the real class..
        Node<OWLClass> topclass = reasoner.getTopClassNode();
        OWLClass thing = topclass.getRepresentativeElement();
        System.out.println("things: " + reasoner.getInstances(thing, false));

        for (OWLClass owlClass : reasoner.getTopClassNode())
            System.out.println(owlClass);

        System.out.println("TopClass: " + thing);
        System.out.println(thing.getIRI());
        System.out.println(thing.getIndividuals(ontology));

        NodeSet<OWLNamedIndividual> individuals = reasoner.getInstances(position, false);
        for (Node<OWLNamedIndividual> sameInd : individuals) {
            System.out.println("Found Individual");
        }

        // we create an empty ontology with namespaces and perform some queries
        for (Role r : reasoner.getKB().getRBox().getRoles())
            System.out.println(r.getName() + " " + r.getDomains() + " " + r.getType() + " " + r.getRanges() + " ");

        System.out.println(reasoner.getKB().getABoxAssertions(KnowledgeBase.AssertionType.TYPE));
        System.out.println(reasoner.getKB().getABoxAssertions(KnowledgeBase.AssertionType.OBJ_ROLE));
        System.out.println(reasoner.getKB().getABoxAssertions(KnowledgeBase.AssertionType.DATA_ROLE));
        System.out.println(reasoner.getKB().getObjectProperties());
        org.semanticweb.owlapi.reasoner.impl.OWLNamedIndividualNode node = (OWLNamedIndividualNode) reasoner.getInstances(factory.getOWLClass(IRI.create(Builtin.namespace() + "WheelChair")), false).iterator().next();
        System.out.println(node.getRepresentativeElement().getObjectPropertyValues(ontology));
        System.out.println(node.getRepresentativeElement().getDataPropertyValues(ontology));
        System.out.println(reasoner.getKB().getObjectPropertyValues(ATermUtils.makeTermAppl(Builtin.namespace() + "at"), ATermUtils.makeTermAppl(Builtin.namespace() + "rolland")));
        System.out.println(reasoner.getKB().getObjectPropertyValues(ATermUtils.makeTermAppl(Builtin.namespace() + "at"), ATermUtils.makeTermAppl(Builtin.namespace() + "rolland2")));
        System.out.println(reasoner.getKB().getTBox().getAxioms());
        System.out.println(reasoner.getKB().getTBox().getAssertedAxioms());
    }

    private static void printABox(PelletReasoner reasoner) {
        System.out.println("=== ABOX ===");
        for (org.mindswap.pellet.Node n : reasoner.getKB().getABox().getNodes()) {
            System.out.println(n);
            System.out.println(n.getName());
        }
        System.out.println("======");
    }

    private static void printTBox(PelletReasoner reasoner) {
        System.out.println("=== TBOX ===");
        for (ATerm term : reasoner.getKB().getTBox().getAxioms()) {
            System.out.println(term);
        }
        System.out.println("======");
    }

// --Commented out by Inspection START (05.07.13 18:26):
//    private static void addClass(PelletReasoner reasoner) {
//        ATermFactory aTermFactory = new aterm.pure.PureFactory();
//        AFun myterm = aTermFactory.makeAFun(Builtin.namespace() + "Test", 0, false);
//
//        Set<ATermAppl> classes = reasoner.getKB().getClasses();
//        Iterator it = classes.iterator();
//        ATermAppl superClass = (ATermAppl) it.next();
//
//        ATermAppl appl = aTermFactory.makeAppl(myterm);
//        reasoner.getKB().addClass(appl);
//
//        reasoner.getKB().addSubClass(appl, superClass);
//    }
// --Commented out by Inspection STOP (05.07.13 18:26)

    private static void printRBox(PelletReasoner reasoner) {
        System.out.println("=== RBOX ===");
        for (ATerm term : reasoner.getKB().getRBox().getRoleNames()) {
            System.out.println(term);
        }
        System.out.println("======");
    }

    private static void printClasses(PelletReasoner reasoner) {
        System.out.println("=== Classes ===");
        for (ATerm term : reasoner.getKB().getAllClasses()) {
            System.out.println(term);
            if (term instanceof ATermAppl) {
                AFun fun = ((ATermAppl) term).getAFun();
                System.out.println(fun.getName());
                System.out.println(term);
            }
        }
        System.out.println("======");
    }
}
