package src.test

import com.clarkparsia.pellet.owlapiv3.PelletReasoner
import de.derivo.sparqldlapi.{Query, QueryResult, QueryEngine}
import java.io._
import org.semanticweb.owlapi.apibinding.OWLManager
import org.semanticweb.owlapi.model.OWLOntologyManager
import org.semanticweb.owlapi.reasoner.structural.StructuralReasonerFactory
import src.ship._
import src.ship.external.Ontology2Scala

/**
 * Created by IntelliJ IDEA.
 * User: dodi
 * Date: 7/23/12
 * Time: 10:21 AM
 */

object SparQLTest extends OntologyParser {
  def main(args: Array[String]) {
    // Create an ontology manager
    val manager: OWLOntologyManager = OWLManager.createOWLOntologyManager()
    // Load the wine ontology from the web.

    //val ont : OWLOntology = manager.loadOntologyFromOntologyDocument(IRI.create("http://www.w3.org/TR/owl-guide/wine.rdf"));

    // Create an instance of an OWL API reasoner (we use the OWL API built-in StructuralReasoner for the purpose of demonstration here)
    val factory: StructuralReasonerFactory = new StructuralReasonerFactory()


    parseAll(ontology, new FileReader(new File("examples/baall/transport6.ont"))) match {
      case Success(ont, _) =>
        val pm = Builtin.prefixManager
        val owlont = Ontology2OWL.convert(ont, pm, manager)
        //Ontology2OWL.addComplementClasses(owlont, Ontology2OWL.computeTopClasses(owlont), manager, manager.getOWLDataFactory)
        //Ontology2OWL.addDistinctnessAxioms(owlont, Ontology2OWL.computeTopClasses(owlont), manager)

        manager.saveOntology(owlont, new FileOutputStream(new File("examples/transport.owl")))

        // Optionally let the reasoner compute the most relevant inferences in advance
        //reasoner.precomputeInferences(InferenceType.CLASS_ASSERTIONS,InferenceType.OBJECT_PROPERTY_ASSERTIONS);
        val reasoner2 = new PelletReasoner(owlont, org.semanticweb.owlapi.reasoner.BufferingMode.NON_BUFFERING)
        reasoner2.prepareReasoner()
        reasoner2.getKB.realize()

        //val internSet1 = ATermUtils.makeTermAppl("http://www.semanticweb.org/ontologies/2012/6/SHIP.owl#Free");
        //println("instances of " + internSet1 + " " + reasoner2.getKB.getInstances(internSet1, false))

        // Create an instance of the SPARQL-DL query engine
        val engine: QueryEngine = QueryEngine.create(manager, reasoner2, true)

        var squery: Query = Query.create("PREFIX ship: <" + pm.getDefaultPrefix + ">\n" +
          "SELECT * WHERE {\n"
          + " PropertyValue(?subject, ship:at, ?object)}")
        println(squery)
        var result: QueryResult = engine.execute(squery)
        println("res1: " + result)
        var squery2: Query = OWLQuery.aBoxFormulas2Query(List[ABoxFormula](RoleAssertion(Var("x"), Var("y"), NamedRole("at"))), pm)
        println(squery2)
        var result2: QueryResult = engine.execute(squery2)
        println("res2: " + result2)
        println("res3: " + Ontology2Scala.RoleEntriesAsList(ont, NamedRole("neighbour")))
      case NoSuccess(msg, _) => println("parse failure: " + msg)
    }
  }
}
