package src.test;

/**
 * Created by IntelliJ IDEA.
 * User: dodi
 * Date: 7/18/12
 * Time: 2:06 PM
 * To change this template use File | Settings | File Templates.
 */


import com.clarkparsia.pellet.owlapiv3.PelletReasonerFactory;
import org.semanticweb.owlapi.apibinding.OWLManager;
import org.semanticweb.owlapi.model.*;
import org.semanticweb.owlapi.reasoner.*;
import org.semanticweb.owlapi.util.DefaultPrefixManager;

import java.io.File;
import java.io.FileOutputStream;

/**
 * Author: Matthew Horridge<br>
 * The University of Manchester<br>
 * Information Management Group<br>
 * Date: 18-Dec-2009
 * </p>
 * This example shows how to get access to objects that represent entities.
 */
public class Test3 {

    public static void main(String[] args) {

        try {
            // In order to get access to objects that represent entities we need a data factory.
            OWLOntologyManager manager = OWLManager.createOWLOntologyManager();

            // We can get a reference to a data factory from an OWLOntologyManager.
            OWLDataFactory factory = manager.getOWLDataFactory();

            // In OWL, entities are named objects that are used to build class expressions and axioms.  They
            // include classes, properties (object, data and annotation), named individuals and datatypes.
            // All entities may be obtained from an OWLDataFactory.

            // Let's create an object to represent a class.  In this case, we'll choose
            // http://www.semanticweb.org/owlapi/ontologies/ontology#A
            // as the IRI for our class.

            // There are two ways we can create classes (and other entities).


            // The second is to use a prefix manager and specify abbreviated IRIs.  This is useful for creating
            // lots of entities with the same prefix IRIs.
            // First create our prefix manager and specify that the default prefix IRI (bound to the empty prefix name)
            // is http://www.semanticweb.org/owlapi/ontologies/ontology#
            PrefixManager pm = new DefaultPrefixManager("http://www.semanticweb.org/owlapi/ontologies/ontology#");
            // Now we use the prefix manager and just specify an abbreviated IRI


            // Note that clsAMethodA will be equal to clsAMethodB because they are both OWLClass objects and have the
            // same IRI.

            // Creating entities in the above manner does not "add them to an ontology".  They are merely objects that
            // allow us to reference certain objects (classes etc.) for use in class expressions, and axioms (which can
            // be added to an ontology).

            // Lets create an ontology, and add a declaration axiom to the ontology that declares the above class
            OWLOntology ontology = manager.createOntology(IRI.create("http://www.semanticweb.org/owlapi/ontologies/ontology"));
            // We can add a declaration axiom to the ontology, that essentially adds the class to the signature of



            // Note that it isn't necessary to add declarations to an ontology in order to use an entity.  For some
            // ontology formats (e.g. the Manchester Syntax), declarations will automatically be added in the saved
            // version of the ontology.


            OWLDataFactory dataFactory = manager.getOWLDataFactory();
            // Get the reference to the :Person class (the full IRI will be <http://example.com/owl/families/Person>)
            OWLClass person = dataFactory.getOWLClass(":Person", pm);
            OWLDeclarationAxiom clsDeclPerson = factory.getOWLDeclarationAxiom(person);
            manager.addAxiom(ontology, clsDeclPerson);

            // Get the reference to the :Mary class (the full IRI will be <http://example.com/owl/families/Mary>)
            OWLNamedIndividual mary = dataFactory.getOWLNamedIndividual(":mary", pm);
            OWLNamedIndividual john = dataFactory.getOWLNamedIndividual(":john", pm);

            // Now create a ClassAssertion to specify that :Mary is an instance of :Person
            OWLClassAssertionAxiom maryPerson = dataFactory.getOWLClassAssertionAxiom(person, mary);
            manager.addAxiom(ontology, maryPerson);
            OWLClassAssertionAxiom johnPerson = dataFactory.getOWLClassAssertionAxiom(person, john);
            manager.addAxiom(ontology, johnPerson);

            OWLClass clsA = factory.getOWLClass(":A",pm);
            OWLClass clsB = factory.getOWLClass(":B",pm);

            // Now create the axiom
            OWLAxiom subclassaxiom = factory.getOWLSubClassOfAxiom(clsA, clsB);
            manager.addAxiom(ontology, subclassaxiom);

            OWLObjectProperty hasFather = factory.getOWLObjectProperty(":hasFather",pm);
            OWLObjectPropertyAssertionAxiom fatheraxiom = dataFactory.getOWLObjectPropertyAssertionAxiom(hasFather, mary, john);
            manager.addAxiom(ontology, fatheraxiom);

            manager.addAxiom(ontology, dataFactory.getOWLObjectPropertyDomainAxiom(hasFather, person));
            manager.addAxiom(ontology, dataFactory.getOWLObjectPropertyRangeAxiom(hasFather, person));

            OWLClass clsChild = factory.getOWLClass(":Child", pm);


            manager.addAxiom(ontology, dataFactory.getOWLEquivalentClassesAxiom(clsChild,
                    dataFactory.getOWLObjectSomeValuesFrom(hasFather, person)
            ));


            OWLReasonerFactory reasonerFactory = new PelletReasonerFactory();
            // OWLReasonerFactory reasonerFactory = new Reasoner.ReasonerFactory();

            // We'll now create an instance of an OWLReasoner (the implementation being provided by HermiT as
            // we're using the HermiT reasoner factory). The are two categories of reasoner, Buffering and
            // NonBuffering. In our case, we'll create the buffering reasoner, which is the default kind of reasoner.
            // We'll also attach a progress monitor to the reasoner. To do this we set up a configuration that
            // knows about a progress monitor.

            // Create a console progress monitor. This will print the reasoner progress out to the console.
            ConsoleProgressMonitor progressMonitor = new ConsoleProgressMonitor();
            // Specify the progress monitor via a configuration. We could also specify other setup parameters in
            // the configuration, and different reasoners may accept their own defined parameters this way.
            OWLReasonerConfiguration config = new SimpleConfiguration(progressMonitor);
            // Create a reasoner that will reason over our ontology and its imports closure. Pass in the configuration.

            OWLReasoner reasoner = reasonerFactory.createReasoner(ontology, config);
            System.out.println("Childs: " + reasoner.getInstances(clsChild, false));
            System.out.println("Person: "+reasoner.getInstances(person, false));



            try {
                FileOutputStream fos = new FileOutputStream(new File("/tmp/test.owl"));
                manager.saveOntology(ontology, fos);
            } catch (Exception e) {
                e.printStackTrace();  //To change body of catch statement use File | Settings | File Templates.
            }

        }
        catch (OWLOntologyCreationException e) {
            System.out.println("Could not create ontology: " + e.getMessage());
        }

    }
}
